# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

import collections.abc
import io
import logging
import os
import sys
import threading
import time

from pyomo.common.errors import DeveloperError
from pyomo.common.log import LoggingIntercept, LogStream

_poll_interval = 0.0001
_poll_rampup_limit = 0.099
# reader polling: number of timeouts with no data before increasing the
# polling interval
_poll_rampup = 10
# polling timeout when waiting to close threads.  This will bail on
# closing threast after a minimum of 13.1 seconds and a worst case of
# ~(13.1 * #threads) seconds
_poll_timeout = 1  # 14 rounds: 0.0001 * 2**14 == 1.6384
_poll_timeout_deadlock = 100  # seconds
_pipe_buffersize = 1 << 16  # 65536
_noop = lambda: None
_mswindows = sys.platform.startswith('win')
try:
    if _mswindows:
        from msvcrt import get_osfhandle
        from win32file import ReadFile
        from win32pipe import FdCreatePipe, PeekNamedPipe, SetNamedPipeHandleState

        # This constant from Microsoft SetNamedPipeHandleState documentation:
        PIPE_NOWAIT = 1
    else:
        from select import select
    _peek_available = True
except ImportError:
    _peek_available = False

logger = logging.getLogger(__name__)


class _SignalFlush:
    def __init__(self, ostream, handle):
        super().__setattr__('_ostream', ostream)
        super().__setattr__('_handle', handle)

    if _mswindows:
        # Because we are setting the pipe to be non-blocking in Windows,
        # it is possible that calls to flush() and write() will raise
        # BlockingIOError.  We will catch and retry.  In addition, we
        # will chunk the data into pieces that should be well below the
        # pipe buffer size (so we should avoid deadlock)

        def _retry(self, fcn, *args, retries=10):
            # Attempting to write to the pipe in the testing harness
            # occasionally raises OSError ("No space left on disk") or
            # BlockingIOError (when the write would be truncated).  We
            # will re-try after a brief pause.
            failCount = 0
            while 1:
                try:
                    return fcn(*args)
                except (OSError, BlockingIOError):
                    failCount += 1
                    if failCount >= retries:
                        raise
                    time.sleep(_poll_rampup_limit / (retries - 1))

        def flush(self):
            self._retry(self._ostream.flush)
            self._handle.flush = True

        def write(self, data: str) -> int:
            ans = 0
            chunksize = _pipe_buffersize >> 1  # 1/2 the buffer size
            for i in range(0, len(data), chunksize):
                ans += self._retry(self._ostream.write, data[i : i + chunksize])
            return ans

        def writelines(self, data):
            for line in data:
                self.write(line)

    else:

        def flush(self):
            self._ostream.flush()
            self._handle.flush = True

    def __getattr__(self, attr):
        return getattr(self._ostream, attr)

    def __setattr__(self, attr, val):
        return setattr(self._ostream, attr, val)


class _AutoFlush(_SignalFlush):
    if _mswindows:
        # Because we define write() and writelines() under windows, we
        # need to make sure that _AutoFlush calls them

        def write(self, data: str) -> int:
            ans = super().write(data)
            self.flush()
            return ans

        def writelines(self, data):
            super().writelines(data)
            self.flush()

    else:

        def write(self, data: str) -> int:
            ans = self._ostream.write(data)
            self.flush()
            return ans

        def writelines(self, data):
            self._ostream.writelines(data)
            self.flush()


class _fd_closer:
    """A context manager to handle closing a specified file descriptor

    Ideally we would use `os.fdopen(... closefd=True)`; however, it
    appears that Python ignores `closefd` on Windows.  This would
    eventually lead to the process exceeding the maximum number of open
    files (see Pyomo/pyomo#3587).  So, we will explicitly manage closing
    the file descriptors that we open using this context manager.

    """

    def __init__(self, fd):
        self.fd = fd

    def __enter__(self):
        return self.fd

    def __exit__(self, et, ev, tb):
        os.close(self.fd)


class redirect_fd:
    """Redirect a file descriptor to a new file or file descriptor.

    This context manager will redirect the specified file descriptor to
    a specified new output target (either file name or file descriptor).
    For the special case of file descriptors 1 (stdout) and 2 (stderr),
    we will also make sure that the Python `sys.stdout` or `sys.stderr`
    remain usable: in the case of synchronize=True, the `sys.stdout` /
    `sys.stderr` file handles point to the new file descriptor.  When
    synchronize=False, we preserve the behavior of the Python file
    object (retargeting it to the original file descriptor if necessary).

    Parameters
    ----------
    fd: int
        The file descriptor to redirect

    output: int or str or None
        The new output target for `fd`: either another valid file
        descriptor (int) or a string with the file to open.  If `None`,
        then the fd is redirected to `os.devnull`.

    synchronize: bool
        If True, and `fd` is 1 or 2, then update `sys.stdout` or
        `sys.stderr` to also point to the new file descriptor

    """

    def __init__(self, fd=1, output=None, synchronize=True):
        if output is None:
            # /dev/null is used just to discard what is being printed
            output = os.devnull
        self.fd = fd
        self.std = {1: 'stdout', 2: 'stderr'}.get(self.fd, None)
        self.target = output
        self.target_file = None
        self.synchronize = synchronize
        self.original_file = None
        self.original_fd = None

    def __enter__(self):
        if self.std:
            # We used to flush original_file here.  We have removed that
            # because the std* streams are flushed by capture_output.
            # Flushing again here caused intermittent errors due to
            # closed file handles on OSX
            self.original_file = getattr(sys, self.std)
        # Duplicate the original standard file descriptor(file
        # descriptor 1 or 2) to a different file descriptor number
        self.original_fd = os.dup(self.fd)

        # Open a file descriptor pointing to the new file
        if isinstance(self.target, int):
            out_fd = self.target
        else:
            out_fd = os.open(self.target, os.O_WRONLY)

        # Duplicate the file descriptor for the opened file, closing and
        # overwriting/replacing the original fd.  Only make the new FD
        # inheritable if it is stdout/stderr
        os.dup2(out_fd, self.fd, inheritable=bool(self.std))

        # We no longer need the original file descriptor
        if not isinstance(self.target, int):
            os.close(out_fd)

        if self.synchronize and self.std:
            # Cause Python's stdout to point to our new file
            self.target_file = os.fdopen(self.fd, "w", closefd=False)
            setattr(sys, self.std, self.target_file)

        return self

    def __exit__(self, et, ev, tb):
        # Close output: this either closes the new file that we opened,
        # or else the new file that points to the original (duplicated)
        # file descriptor
        if self.target_file is not None:
            self.target_file.flush()
            self.target_file.close()
            self.target_file = None
            setattr(sys, self.std, self.original_file)
        # Restore stdout's FD (implicitly closing the FD we opened)
        os.dup2(self.original_fd, self.fd, inheritable=bool(self.std))
        # Close the temporary FD
        os.close(self.original_fd)


class capture_output:
    """Context manager to capture output sent to sys.stdout and sys.stderr

    This is a drop-in substitute for PyUtilib's capture_output to
    temporarily redirect output to the provided stream or file.

    Parameters
    ----------
    output : io.TextIOBase, Sequence[io.TextIOBase], TeeStream, str, or None

        Output stream where all captured stdout/stderr data is sent.  If
        a ``str`` is provided, it is used as a file name and opened
        (potentially overwriting any existing file).  If ``None``, a
        :class:`io.StringIO` object is created and used.

    capture_fd : bool

        If True, we will also redirect the process file descriptors
        ``1`` (stdout), ``2`` (stderr), and the file descriptors from
        ``sys.stdout.fileno()`` and ``sys.stderr.fileno()`` to the
        ``output``.  This is useful for capturing output emitted
        directly to the process stdout / stderr by external compiled
        modules.

        Capturing and redirecting the file descriptors can cause loops
        in the output stream (where one of the `output` streams points
        to a file descriptor we just captured).
        :py:class:`capture_output` will attempt to locate
        :py:class:`io.IOBase` streams in `output` that point to file
        descriptors that we just captured and replace them with
        temporary streams that point to (copies of) the original file
        descriptor.  In addition, :py:class:`capture_output` will look
        for :py:class:`~pyomo.common.log.LogStream` objects and will
        attempt to locate :py:class:`logging.StreamHandle` objects that
        would output to a redirected file descriptor and temporarily
        redirect those handlers to (copies of) the original file
        descriptor.

        Note that this process will cover the most common cases, but is
        by no means perfect.  Use of other output classes or customized
        log handlers may still result in output loops (usually
        manifesting in an error message about text being left in the
        output buffer).

    Returns
    -------
    io.TextIOBase

        This is the output stream object where all data is sent.

    """

    startup_shutdown = threading.Lock()

    def __init__(self, output=None, capture_fd=False):
        self.output = output
        self.output_stream = None
        self.old = None
        self.tee = None
        self.capture_fd = capture_fd
        self.context_stack = []

    def _enter_context(self, cm, prior_to=None):
        """Add the context manager to the context stack and return the result
        from calling the context manager's `__enter__()`

        """
        if prior_to is None:
            self.context_stack.append(cm)
        else:
            self.context_stack.insert(self.context_stack.index(prior_to), cm)
        return cm.__enter__()

    def _exit_context_stack(self, et, ev, tb):
        """Flush the context stack, calling __exit__() on all context managers

        One would usually use the contextlib.ExitStack to implement/manage
        the collection of context managers we are putting together.  The
        problem is that ExitStack will only call the __exit__ handlers
        up to the first one that returns an exception.  As we are
        expecting the possibility of one of the CMs here to raise an
        exception (usually from TeeStream when joining the reader
        threads), we will explicitly implement the stack management here
        so that we will guarantee that all __exit__ handlers will always
        be called.

        """
        FAIL = []
        while self.context_stack:
            try:
                cm = self.context_stack.pop()
                cm.__exit__(et, ev, tb)
            except:
                _stack = self.context_stack
                FAIL.append(
                    f"{sys.exc_info()[0].__name__}: {sys.exc_info()[1]} ({len(_stack)+1}: {cm}@{id(cm):x})"
                )
        return FAIL

    def __enter__(self):
        if not capture_output.startup_shutdown.acquire(timeout=_poll_timeout_deadlock):
            # This situation *shouldn't* happen.  If it does, it is
            # unlikely that the user can fix it (or even debug it).
            # Instead they should report it back to us.
            #
            # Breadcrumbs:
            #
            #   - The last time we hit this [5/2025], it was because we
            #     were using capture_output in a solver's __del__.  This
            #     led to the GC deleting the solver while another solver
            #     was trying to start up / run (so the other solver held
            #     the lock, but the GC interrupted that thread and
            #     wouldn't let go).
            raise DeveloperError("Deadlock starting capture_output")
        try:
            return self._enter_impl()
        finally:
            capture_output.startup_shutdown.release()

    def __exit__(self, et, ev, tb):
        if not capture_output.startup_shutdown.acquire(timeout=_poll_timeout_deadlock):
            # See comments & breadcrumbs in __enter__() above.
            raise DeveloperError("Deadlock closing capture_output")
        try:
            return self._exit_impl(et, ev, tb)
        finally:
            capture_output.startup_shutdown.release()

    def _enter_impl(self):
        self.old = (sys.stdout, sys.stderr)
        old_fd = []
        for stream in self.old:
            try:
                stream.flush()
                try:
                    old_fd.append(stream.fileno())
                except (AttributeError, OSError):
                    old_fd.append(None)
            except (ValueError, OSError):
                old_fd.append(None)
        try:
            # We have an issue where we are (very aggressively)
            # commandeering the terminal.  This is what we intend, but the
            # side effect is that any errors generated by this module (e.g.,
            # because the user gave us an invalid output stream) get
            # completely suppressed.  So, we will make an exception to the
            # output that we are catching and let messages logged to THIS
            # logger to still be emitted to the original stderr.
            if self.capture_fd:
                # Because we are also commandeering the FD that underlies
                # sys.stderr, we cannot just write to that stream and
                # instead will open a new stream to the "original" FD
                # (Note that we need to duplicate that FD, as we will
                # overwrite it when we get to redirect_fd below).  If
                # sys.stderr doesn't have a file descriptor, we will
                # fall back on the process stderr (FD=2).
                #
                # Note that we would like to use closefd=True, but can't
                # (see _fd_closer docs)
                log_stream = self._enter_context(
                    os.fdopen(
                        self._enter_context(_fd_closer(os.dup(old_fd[1] or 2))),
                        mode="w",
                        closefd=False,
                    )
                )
            else:
                log_stream = self.old[1]
            self._enter_context(LoggingIntercept(log_stream, logger=logger, level=None))

            if isinstance(self.output, str):
                self.output_stream = self._enter_context(open(self.output, "a"))
            elif self.output is None:
                self.output_stream = io.StringIO()
            else:
                self.output_stream = self.output
            if isinstance(self.output, TeeStream):
                self.tee = self._enter_context(self.output)
            elif isinstance(self.output_stream, collections.abc.Sequence):
                self.tee = self._enter_context(TeeStream(*self.output_stream))
            else:
                self.tee = self._enter_context(TeeStream(self.output_stream))
            fd_redirect = {}
            if self.capture_fd:
                tee_fd = (self.tee.STDOUT.fileno(), self.tee.STDERR.fileno())
                for i in range(2):
                    # Redirect the standard process file descriptor (1 or 2)
                    fd_redirect[i + 1] = self._enter_context(
                        redirect_fd(i + 1, tee_fd[i], synchronize=False)
                    )
                    # Redirect the file descriptor currently associated with
                    # sys.stdout / sys.stderr
                    fd = old_fd[i]
                    if fd and fd not in fd_redirect:
                        fd_redirect[fd] = self._enter_context(
                            redirect_fd(fd, tee_fd[i], synchronize=False)
                        )
            # We need to make sure that we didn't just capture the FD
            # that underlies a stream that we are outputting to.  Note
            # that when capture_fd==False, normal streams will be left
            # alone, but the lastResort _StderrHandler() will still be
            # replaced (needed because that handler uses the *current*
            # value of sys.stderr)
            ostreams = []
            for stream in self.tee.ostreams:
                if isinstance(stream, LogStream):
                    for handler_redirect in stream.redirect_streams(fd_redirect):
                        self._enter_context(handler_redirect, prior_to=self.tee)
                else:
                    try:
                        fd = stream.fileno()
                    except (AttributeError, OSError):
                        fd = None
                    if fd in fd_redirect:
                        # We just redirected this file descriptor so
                        # we can capture the output.  This makes a
                        # loop that we really want to break.  Undo
                        # the redirect by pointing our output stream
                        # back to the original file descriptor.
                        #
                        # Note that we would like to use closefd=True, but can't
                        # (see _fd_closer docs)
                        stream = self._enter_context(
                            os.fdopen(
                                self._enter_context(
                                    _fd_closer(os.dup(fd_redirect[fd].original_fd)),
                                    prior_to=self.tee,
                                ),
                                mode="w",
                                closefd=False,
                            ),
                            prior_to=self.tee,
                        )
                ostreams.append(stream)
            self.tee.ostreams = ostreams
        except:
            # Note: we will ignore any exceptions raised while exiting
            # the context managers and just reraise the original
            # exception.
            self._exit_context_stack(*sys.exc_info())
            raise
        sys.stdout = self.tee.STDOUT
        sys.stderr = self.tee.STDERR
        buf = self.tee.ostreams
        if len(buf) == 1:
            buf = buf[0]
        return buf

    def _exit_impl(self, et, ev, tb):
        # Check that we were nested correctly
        FAIL = []
        if self.tee is not None:
            if self.tee._stdout is not None and self.tee.STDOUT is not sys.stdout:
                FAIL.append(
                    'Captured output (%s) does not match sys.stdout (%s).'
                    % (self.tee._stdout, sys.stdout)
                )
            if self.tee._stderr is not None and self.tee.STDERR is not sys.stderr:
                FAIL.append(
                    'Captured output (%s) does not match sys.stderr (%s).'
                    % (self.tee._stdout, sys.stdout)
                )
        # Exit all context managers.  This includes
        #  - Restore any file descriptors we commandeered
        #  - Close / join the TeeStream
        #  - Close any opened files
        FAIL.extend(self._exit_context_stack(et, ev, tb))
        if self.old is not None:
            sys.stdout, sys.stderr = self.old
            self.old = None
        self.tee = None
        self.output_stream = None
        if FAIL:
            raise RuntimeError("\n".join(FAIL))

    def __del__(self):
        if self.tee is not None:
            self.__exit__(None, None, None)

    def setup(self):
        if self.old is not None:
            raise RuntimeError('Duplicate call to capture_output.setup.')
        return self.__enter__()

    def reset(self):
        return self.__exit__(None, None, None)


class _StreamHandle:
    """A stream handler object used by TeeStream

    This handler holds the two sides of the pipe used to communicate
    output generated by the main thread out to the handler thread (which
    passes the output on to the TeeStream's output streams).

    Note that this class is intimately tied to TeeStream and relies on
    that class for certain termination / cleanup events (including
    flushing and closing buffers)

    """

    def __init__(self, mode, buffering, encoding, newline):
        self.buffering = buffering
        self.newlines = newline
        self.flush = False
        if _peek_available and _mswindows:
            # This is a re-implementation of os.pipe() on Windows so
            # that we can explicitly request a larger pipe buffer (64k;
            # matching *NIX).  Per the docs: on Windows, the pipe buffer
            # should automatically grow if needed.  However, we have
            # observed (see #3658) that if it happens, it can cause
            # deadlock when clients write directly to the underlying
            # file descriptor.  By explicitly requesting a larger buffer
            # from the outset, we reduce the likelihood of needing to
            # reallocate the buffer.
            self.read_pipe, self.write_pipe = FdCreatePipe(
                None, _pipe_buffersize, os.O_BINARY if 'b' in mode else os.O_TEXT
            )
            self.read_pyhandle = get_osfhandle(self.read_pipe)
            self.write_pyhandle = get_osfhandle(self.write_pipe)
            # Because reallocating the pipe buffer can cause deadlock
            # (at least in the context in which we are using pipes
            # here), we will set the write pipe to NOWAIT.  This will
            # guarantee that we don't deadlock, but at the cost of
            # possibly losing some output (the fprintf() to the FD will
            # return a number of bytes written less than the string that
            # was passed.  If the client is ignoring the return value,
            # then *poof*: the output is truncated)
            SetNamedPipeHandleState(self.write_pyhandle, PIPE_NOWAIT, None, None)
        else:
            self.read_pipe, self.write_pipe = os.pipe()

        if not buffering and 'b' not in mode:
            # While we support "unbuffered" behavior in text mode,
            # python does not
            buffering = -1
        self.write_file = os.fdopen(
            self.write_pipe,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            newline=newline,
            closefd=False,
        )
        if not self.buffering and buffering:
            # We want this stream to be unbuffered, but Python doesn't
            # allow it for text streams.  Mock up an unbuffered stream
            # using AutoFlush
            self.write_file = _AutoFlush(self.write_file, self)
        else:
            self.write_file = _SignalFlush(self.write_file, self)
        self.decoder_buffer = b''
        try:
            self.encoding = encoding or self.write_file.encoding
        except AttributeError:
            self.encoding = None
        if self.encoding:
            self.output_buffer = ''
        else:
            self.output_buffer = b''

    def fileno(self):
        return self.read_pipe

    def close(self):
        # Close both the file and the underlying file descriptor.  Note
        # that this may get called more than once.
        if self.write_file is not None:
            if not self.write_file.closed:
                self.write_file.flush()
                self.write_file.close()
            self.write_file = None

        if self.write_pipe is not None:
            # Close the write side of the pipe: the reader thread is
            # waiting for the EOF so that it can shut down.
            try:
                # If someone else has closed the file descriptor, then
                # python raises an OSError
                os.close(self.write_pipe)
            except OSError:
                pass
            self.write_pipe = None

    def finalize(self, ostreams):
        # Note that this expects to be called by TeeStream *after*
        # TeeStream has called close(), so the output_buffer should have
        # been flushed and emptied.
        self.decodeIncomingBuffer()
        if ostreams:
            self.writeOutputBuffer(ostreams, True)
        # Close the read side of the pipe.
        os.close(self.read_pipe)
        if self.decoder_buffer:
            logger.error(
                "Stream handle closed with un-decoded characters "
                "in the decoder buffer that was not emitted to the "
                "output stream(s):\n\t%r" % (self.decoder_buffer,)
            )

    def decodeIncomingBuffer(self):
        if not self.encoding:
            self.output_buffer, self.decoder_buffer = self.decoder_buffer, b''
            return

        raw_len = len(self.decoder_buffer)
        chars = ''
        while raw_len:
            try:
                chars = self.decoder_buffer[:raw_len].decode(self.encoding)
                break
            except:
                pass
            # partial read of unicode character, try again with
            # a shorter bytes buffer
            raw_len -= 1
        if self.newlines is None:
            chars = chars.replace('\r\n', '\n').replace('\r', '\n')
        self.output_buffer += chars
        self.decoder_buffer = self.decoder_buffer[raw_len:]

    def writeOutputBuffer(self, ostreams, flush):
        if not self.encoding:
            ostring, self.output_buffer = self.output_buffer, b''
        elif self.buffering > 0 and not flush:
            EOL = self.output_buffer.rfind(self.newlines or '\n') + 1
            ostring = self.output_buffer[:EOL]
            self.output_buffer = self.output_buffer[EOL:]
        else:
            ostring, self.output_buffer = self.output_buffer, ''

        if not ostring:
            return

        for stream in ostreams:
            try:
                written = stream.write(ostring)
            except:
                my_repr = "<%s.%s @ %s>" % (
                    stream.__class__.__module__,
                    stream.__class__.__name__,
                    hex(id(stream)),
                )
                if my_repr in ostring:
                    # In the case of nested capture_outputs, all the
                    # handlers are left on the logger.  We want to make
                    # sure that we don't create an infinite loop by
                    # re-printing a message *this* object generated.
                    continue
                et, e, tb = sys.exc_info()
                msg = "Error writing to output stream %s:\n    %s: %s\n" % (
                    my_repr,
                    et.__name__,
                    e,
                )
                if getattr(stream, 'closed', False):
                    msg += "Output stream closed before all output was written to it."
                else:
                    msg += "Is this a writeable TextIOBase object?"
                logger.error(
                    f"{msg}\nThe following was left in the output buffer:\n"
                    f"    {ostring!r}"
                )
                continue
            if flush or (written and not self.buffering):
                getattr(stream, 'flush', _noop)()
            # Note: some derived file-like objects fail to return the
            # number of characters written (and implicitly return None).
            # If we get None, we will just assume that everything was
            # fine (as opposed to tossing the incomplete write error).
            if written is not None and written != len(ostring):
                my_repr = "<%s.%s @ %s>" % (
                    stream.__class__.__module__,
                    stream.__class__.__name__,
                    hex(id(stream)),
                )
                if my_repr in ostring[written:]:
                    continue
                logger.error(
                    "Incomplete write to output stream %s.\nThe following was "
                    "left in the output buffer:\n    %r" % (my_repr, ostring[written:])
                )


class TeeStream:
    def __init__(self, *ostreams, encoding=None, buffering=-1):
        self.ostreams = ostreams
        self.encoding = encoding
        self.buffering = buffering
        self._stdout = None
        self._stderr = None
        self._handles = []
        self._active_handles = []
        self._threads = []
        self._enter_count = 0

    @property
    def STDOUT(self):
        if self._stdout is None:
            b = self.buffering
            if b == -1:
                b = 1
            self._stdout = self.open(buffering=b)
        return self._stdout

    @property
    def STDERR(self):
        if self._stderr is None:
            b = self.buffering
            if b == -1:
                b = 0
            self._stderr = self.open(buffering=b)
        return self._stderr

    def open(self, mode="w", buffering=-1, encoding=None, newline=None):
        if encoding is None:
            encoding = self.encoding
        handle = _StreamHandle(mode, buffering, encoding, newline)
        # Note that is it VERY important to close file handles in the
        # same thread that opens it.  If you don't you can see deadlocks
        # and a peculiar error ("libgcc_s.so.1 must be installed for
        # pthread_cancel to work"; see
        # https://github.com/python/cpython/issues/62948)
        #
        # To accomplish this, we will keep two handle lists: one is the
        # set of "active" handles that the (merged reader) thread is
        # using, and the other the list of all handles so the original
        # thread can close them after the reader thread terminates.
        if handle.buffering:
            self._active_handles.append(handle)
        else:
            # Unbuffered handles should appear earlier in the list so
            # that they get processed first
            self._active_handles.insert(0, handle)
        self._handles.append(handle)
        self._start(handle)
        return handle.write_file

    def close(self, in_exception=False):
        # Close all open handles.  Note that as the threads may
        # immediately start removing handles from the list, it is
        # important that we iterate over a copy of the list.
        for h in list(self._handles):
            h.close()

        # Join all stream processing threads
        _poll = _poll_interval
        _timeout = 0.0
        FAIL = False
        while True:
            for th in self._threads:
                th.join(_poll)
                _timeout += _poll
            self._threads[:] = [th for th in self._threads if th.is_alive()]
            if not self._threads:
                break
            if _poll < _poll_timeout:
                _poll *= 2.0
            if _poll_timeout * 0.5 <= _poll < _poll_timeout:
                if in_exception:
                    # We are already processing an exception: no reason
                    # to trigger another, nor to deadlock for an
                    # extended time.  Silently clean everything up
                    # (because emitting logger messages could trigger
                    # yet another exception and mask the true cause).
                    break
                logger.warning(
                    "Significant delay observed waiting to join reader "
                    "threads, possible output stream deadlock"
                )
            elif _timeout > _poll_timeout_deadlock:
                logger.error("TeeStream: deadlock observed joining reader threads")
                # Defer raising the exception until after we have
                # cleaned things up
                FAIL = True
                break

        for h in list(self._handles):
            h.finalize(self.ostreams)

        self._threads.clear()
        self._handles.clear()
        self._active_handles.clear()
        self._stdout = None
        self._stderr = None
        if FAIL:
            raise RuntimeError("TeeStream: deadlock observed joining reader threads")

    def __enter__(self):
        self._enter_count += 1
        return self

    def __exit__(self, et, ev, tb):
        if not self._enter_count:
            raise RuntimeError("TeeStream: exiting a context that was not entered")
        self._enter_count -= 1
        if not self._enter_count:
            self.close(et is not None)

    def __del__(self):
        # Implement __del__ to guarantee that file descriptors are closed
        # ... but only if we are not called by the GC in a handler thread
        if threading.current_thread() not in self._threads:
            self.close()

    def _start(self, handle):
        if not _peek_available:
            th = threading.Thread(target=self._streamReader, args=(handle,))
            th.daemon = True
            th.start()
            self._threads.append(th)
        elif not self._threads:
            th = threading.Thread(target=self._mergedReader)
            th.daemon = True
            th.start()
            self._threads.append(th)
        else:
            # The merged reader is already running... nothing additional
            # needs to be done
            pass

    def _streamReader(self, handle):
        while True:
            new_data = os.read(handle.read_pipe, io.DEFAULT_BUFFER_SIZE)
            if handle.flush:
                flush = True
                handle.flush = False
            else:
                flush = False
            if new_data:
                handle.decoder_buffer += new_data
            elif not flush:
                break

            # At this point, we have new data sitting in the
            # handle.decoder_buffer
            handle.decodeIncomingBuffer()
            # Now, output whatever we have decoded to the output streams
            handle.writeOutputBuffer(self.ostreams, flush)
        #
        # print("STREAM READER: DONE")

    def _mergedReader(self):
        noop = []
        handles = self._active_handles
        _poll = _poll_interval
        _fast_poll_ct = _poll_rampup
        new_data = ''  # something not None
        while handles:
            flush = False
            if new_data is None:
                # For performance reasons, we use very aggressive
                # polling at the beginning (_poll_interval) and then
                # ramp up to a much more modest polling interval
                # (_poll_rampup_limit) as the process runs and the
                # frequency of new data appearing on the pipe slows
                if _fast_poll_ct:
                    _fast_poll_ct -= 1
                    if not _fast_poll_ct:
                        _poll *= 10
                        if _poll < _poll_rampup_limit:
                            # reset the counter (to potentially increase
                            # the polling interval again)
                            _fast_poll_ct = _poll_rampup
            else:
                new_data = None
            if _mswindows:
                for handle in list(handles):
                    try:
                        pipe = handle.read_pyhandle
                        numAvail = PeekNamedPipe(pipe, 0)[1]
                        if numAvail:
                            result, new_data = ReadFile(pipe, numAvail, None)
                            handle.decoder_buffer += new_data
                            break
                        elif handle.flush:
                            break
                    except:
                        handles.remove(handle)
                        new_data = ''  # not None so the poll interval doesn't increase
                if new_data is None and not handle.flush:
                    # PeekNamedPipe is non-blocking; to avoid swamping
                    # the core, sleep for a "short" amount of time
                    time.sleep(_poll)
                    continue
            else:
                # Because we could be *adding* handles to the TeeStream
                # while the _mergedReader is running, we want to
                # periodically time out and update the list of handles
                # that we are waiting for.  It is also critical that we
                # send select() a *copy* of the handles list, as we see
                # deadlocks when handles are added while select() is
                # waiting
                ready_handles = select(list(handles), noop, noop, _poll)[0]
                if ready_handles:
                    handle = ready_handles[0]
                    new_data = os.read(handle.read_pipe, io.DEFAULT_BUFFER_SIZE)
                    if new_data:
                        handle.decoder_buffer += new_data
                    else:
                        handles.remove(handle)
                        new_data = ''  # not None so the poll interval doesn't increase
                else:
                    for handle in handles:
                        if handle.flush:
                            break
                    else:
                        continue

            if handle.flush:
                new_data = ''
                flush = True
                handle.flush = False

            # At this point, we have new data sitting in the
            # handle.decoder_buffer
            handle.decodeIncomingBuffer()

            # Now, output whatever we have decoded to the output streams
            handle.writeOutputBuffer(self.ostreams, flush)
        #
        # print("MERGED READER: DONE")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.common.collections import ComponentMap
from pyomo.common.deprecation import deprecated
from pyomo.core.kernel.base import ICategorizedObject, _abstract_readonly_property
from pyomo.core.kernel.dict_container import DictContainer
from pyomo.core.kernel.container_utils import define_homogeneous_container_type

logger = logging.getLogger('pyomo.core')

_noarg = object()


# Note: ComponentMap is first in the inheritance chain
#       because its __getstate__ / __setstate__ methods
#       contain some special hacks that allow it to be used
#       for the AML-Suffix object as well (hopefully,
#       temporary). As a result, we need to override the
#       __str__ method on this class so that suffix behaves
#       like ICategorizedObject instead of ComponentMap
class ISuffix(ComponentMap, ICategorizedObject):
    """The interface for suffixes."""

    __slots__ = ()

    #
    # Implementations can choose to define these
    # properties as using __slots__, __dict__, or
    # by overriding the @property method
    #

    direction = _abstract_readonly_property(doc="The suffix direction")
    datatype = _abstract_readonly_property(doc="The suffix datatype")

    #
    # Interface
    #

    def __str__(self):
        return ICategorizedObject.__str__(self)


class suffix(ISuffix):
    """A container for storing extraneous model data that
    can be imported to or exported from a solver."""

    _ctype = ISuffix
    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_direction",
        "_datatype",
        "__weakref__",
    )

    # neither sent to solver or received from solver
    LOCAL = 0
    # sent to solver or other external location
    EXPORT = 1
    # obtained from solver or other external source
    IMPORT = 2
    # both
    IMPORT_EXPORT = 3

    _directions = {
        LOCAL: 'suffix.LOCAL',
        EXPORT: 'suffix.EXPORT',
        IMPORT: 'suffix.IMPORT',
        IMPORT_EXPORT: 'suffix.IMPORT_EXPORT',
    }

    # datatypes (numbers are compatible with ASL bitcodes)
    FLOAT = 4
    INT = 0
    _datatypes = {FLOAT: 'suffix.FLOAT', INT: 'suffix.INT', None: str(None)}

    def __init__(self, *args, **kwds):
        self._parent = None
        self._storage_key = None
        self._active = True
        self._direction = None
        self._datatype = None

        # call the setters
        self.direction = kwds.pop('direction', suffix.LOCAL)
        self.datatype = kwds.pop('datatype', suffix.FLOAT)
        super(suffix, self).__init__(*args, **kwds)

    #
    # Interface
    #

    def export_enabled(self):
        """Returns :const:`True` when this suffix is enabled
        for export to solvers."""
        return bool(self._direction & suffix.EXPORT)

    def import_enabled(self):
        """Returns :const:`True` when this suffix is enabled
        for import from solutions."""
        return bool(self._direction & suffix.IMPORT)

    @property
    def datatype(self):
        """Return the suffix datatype."""
        return self._datatype

    @datatype.setter
    def datatype(self, datatype):
        """Set the suffix datatype."""
        if datatype not in self._datatypes:
            raise ValueError(
                "Suffix datatype must be one of: %s. \n"
                "Value given: %s" % (list(self._datatypes.values()), datatype)
            )
        self._datatype = datatype

    @property
    def direction(self):
        """Return the suffix direction."""
        return self._direction

    @direction.setter
    def direction(self, direction):
        """Set the suffix direction."""
        if not direction in self._directions:
            raise ValueError(
                "Suffix direction must be one of: %s. \n"
                "Value given: %s" % (list(self._directions.values()), direction)
            )
        self._direction = direction

    #
    # Methods that are deprecated
    #

    @deprecated("suffix.set_all_values will be removed in the future.", version='5.3')
    def set_all_values(self, value):
        for ndx in self:
            self[ndx] = value

    @deprecated(
        "suffix.clear_value will be removed in the future. "
        "Use 'del suffix[key]' instead.",
        version='5.3',
    )
    def clear_value(self, component):
        try:
            del self[component]
        except KeyError:
            pass

    @deprecated("suffix.clear_all_values is replaced with suffix.clear", version='5.3')
    def clear_all_values(self):
        self.clear()

    @deprecated(
        "suffix.get_datatype is replaced with the property suffix.datatype",
        version='5.3',
    )
    def get_datatype(self):
        return self.datatype

    @deprecated(
        "suffix.set_datatype is replaced with the property setter suffix.datatype",
        version='5.3',
    )
    def set_datatype(self, datatype):
        self.datatype = datatype

    @deprecated(
        "suffix.get_direction is replaced with the property suffix.direction",
        version='5.3',
    )
    def get_direction(self):
        return self.direction

    @deprecated(
        "suffix.set_direction is replaced with the property setter suffix.direction",
        version='5.3',
    )
    def set_direction(self, direction):
        self.direction = direction


# A list of convenient suffix generators, including:
#   - export_suffix_generator
#       **(used by problem writers)
#   - import_suffix_generator
#       **(used by OptSolver and PyomoModel._load_solution)
#   - local_suffix_generator
#   - suffix_generator


def export_suffix_generator(blk, datatype=_noarg, active=True, descend_into=True):
    """
    Generates an efficient traversal of all suffixes that
    have been declared for exporting data.

    Args:
        blk: A block object.
        datatype: Restricts the suffixes included in the
            returned generator to those matching the
            provided suffix datatype.
        active (:const:`True`/:const:`None`): Controls
            whether or not to filter the iteration to
            include only the active part of the storage
            tree. The default is :const:`True`. Setting this
            keyword to :const:`None` causes the active
            status of objects to be ignored.
        descend_into (bool, function): Indicates whether or
            not to descend into a heterogeneous
            container. Default is True, which is equivalent
            to `lambda x: True`, meaning all heterogeneous
            containers will be descended into.

    Returns:
        iterator of suffixes
    """
    for suf in filter(
        lambda x: (
            x.export_enabled() and ((datatype is _noarg) or (x.datatype is datatype))
        ),
        blk.components(ctype=suffix._ctype, active=active, descend_into=descend_into),
    ):
        yield suf


def import_suffix_generator(blk, datatype=_noarg, active=True, descend_into=True):
    """
    Generates an efficient traversal of all suffixes that
    have been declared for importing data.

    Args:
        blk: A block object.
        datatype: Restricts the suffixes included in the
            returned generator to those matching the
            provided suffix datatype.
        active (:const:`True`/:const:`None`): Controls
            whether or not to filter the iteration to
            include only the active part of the storage
            tree. The default is :const:`True`. Setting this
            keyword to :const:`None` causes the active
            status of objects to be ignored.
        descend_into (bool, function): Indicates whether or
            not to descend into a heterogeneous
            container. Default is True, which is equivalent
            to `lambda x: True`, meaning all heterogeneous
            containers will be descended into.

    Returns:
        iterator of suffixes
    """
    for suf in filter(
        lambda x: (
            x.import_enabled() and ((datatype is _noarg) or (x.datatype is datatype))
        ),
        blk.components(ctype=suffix._ctype, active=active, descend_into=descend_into),
    ):
        yield suf


def local_suffix_generator(blk, datatype=_noarg, active=True, descend_into=True):
    """
    Generates an efficient traversal of all suffixes that
    have been declared local data storage.

    Args:
        blk: A block object.
        datatype: Restricts the suffixes included in the
            returned generator to those matching the
            provided suffix datatype.
        active (:const:`True`/:const:`None`): Controls
            whether or not to filter the iteration to
            include only the active part of the storage
            tree. The default is :const:`True`. Setting this
            keyword to :const:`None` causes the active
            status of objects to be ignored.
        descend_into (bool, function): Indicates whether or
            not to descend into a heterogeneous
            container. Default is True, which is equivalent
            to `lambda x: True`, meaning all heterogeneous
            containers will be descended into.

    Returns:
        iterator of suffixes
    """
    for suf in filter(
        lambda x: (
            x.direction is suffix.LOCAL
            and ((datatype is _noarg) or (x.datatype is datatype))
        ),
        blk.components(ctype=suffix._ctype, active=active, descend_into=descend_into),
    ):
        yield suf


def suffix_generator(blk, datatype=_noarg, active=True, descend_into=True):
    """
    Generates an efficient traversal of all suffixes that
    have been declared.

    Args:
        blk: A block object.
        datatype: Restricts the suffixes included in the
            returned generator to those matching the
            provided suffix datatype.
        active (:const:`True`/:const:`None`): Controls
            whether or not to filter the iteration to
            include only the active part of the storage
            tree. The default is :const:`True`. Setting this
            keyword to :const:`None` causes the active
            status of objects to be ignored.
        descend_into (bool, function): Indicates whether or
            not to descend into a heterogeneous
            container. Default is True, which is equivalent
            to `lambda x: True`, meaning all heterogeneous
            containers will be descended into.

    Returns:
        iterator of suffixes
    """
    for suf in filter(
        lambda x: ((datatype is _noarg) or (x.datatype is datatype)),
        blk.components(ctype=suffix._ctype, active=active, descend_into=descend_into),
    ):
        yield suf


# inserts class definition for simple a
# simple suffix_dict into this module
define_homogeneous_container_type(
    globals(),
    "suffix_dict",
    DictContainer,
    ISuffix,
    doc=("A dict-style container for objects with category type " + ISuffix.__name__),
    use_slots=True,
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.cp import IntervalVar
from pyomo.contrib.cp.scheduling_expr.precedence_expressions import (
    BeforeExpression,
    AtExpression,
)
from pyomo.environ import ConcreteModel, LogicalConstraint, Param


class TestPrecedenceRelationships(unittest.TestCase):
    def get_model(self):
        m = ConcreteModel()
        m.a = IntervalVar()
        m.b = IntervalVar()

        return m

    def test_start_before_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.start_time.before(m.b.start_time))

        self.assertIsInstance(m.c.expr, BeforeExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertEqual(m.c.expr.nargs(), 3)
        self.assertIs(m.c.expr.args[0], m.a.start_time)
        self.assertIs(m.c.expr.args[1], m.b.start_time)
        self.assertEqual(m.c.expr.delay, 0)

        self.assertEqual(str(m.c.expr), "a.start_time <= b.start_time")

    def test_start_before_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.start_time.before(m.b.end_time))

        self.assertIsInstance(m.c.expr, BeforeExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertEqual(m.c.expr.nargs(), 3)
        self.assertIs(m.c.expr.args[0], m.a.start_time)
        self.assertIs(m.c.expr.args[1], m.b.end_time)
        self.assertEqual(m.c.expr.delay, 0)

        self.assertEqual(str(m.c.expr), "a.start_time <= b.end_time")

    def test_start_after_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.start_time.after(m.b.start_time))

        self.assertIsInstance(m.c.expr, BeforeExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertIs(m.c.expr.args[0], m.b.start_time)
        self.assertIs(m.c.expr.args[1], m.a.start_time)
        self.assertEqual(m.c.expr.delay, 0)

        self.assertEqual(str(m.c.expr), "b.start_time <= a.start_time")

    def test_start_after_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.start_time.after(m.b.end_time, delay=2))

        self.assertIsInstance(m.c.expr, BeforeExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertIs(m.c.expr.args[0], m.b.end_time)
        self.assertIs(m.c.expr.args[1], m.a.start_time)
        self.assertEqual(m.c.expr.delay, 2)

        self.assertEqual(str(m.c.expr), "b.end_time + 2 <= a.start_time")

    def test_start_at_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.start_time.at(m.b.start_time))

        self.assertIsInstance(m.c.expr, AtExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertEqual(m.c.expr.nargs(), 3)
        self.assertIs(m.c.expr.args[0], m.a.start_time)
        self.assertIs(m.c.expr.args[1], m.b.start_time)
        self.assertEqual(m.c.expr.delay, 0)

        self.assertEqual(str(m.c.expr), "a.start_time == b.start_time")

    def test_start_at_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.start_time.at(m.b.end_time, delay=-1))

        self.assertIsInstance(m.c.expr, AtExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertEqual(m.c.expr.nargs(), 3)
        self.assertIs(m.c.expr.args[0], m.a.start_time)
        self.assertIs(m.c.expr.args[1], m.b.end_time)
        self.assertEqual(m.c.expr.delay, -1)

        self.assertEqual(str(m.c.expr), "a.start_time - 1 == b.end_time")

    def test_end_before_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.end_time.before(m.b.start_time, delay=3))

        self.assertIsInstance(m.c.expr, BeforeExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertIs(m.c.expr.args[0], m.a.end_time)
        self.assertIs(m.c.expr.args[1], m.b.start_time)
        self.assertEqual(m.c.expr.delay, 3)

        self.assertEqual(str(m.c.expr), "a.end_time + 3 <= b.start_time")

    def test_end_at_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.end_time.at(m.b.start_time, delay=4))

        self.assertIsInstance(m.c.expr, AtExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertIs(m.c.expr.args[0], m.a.end_time)
        self.assertIs(m.c.expr.args[1], m.b.start_time)
        self.assertEqual(m.c.expr.delay, 4)

        self.assertEqual(str(m.c.expr), "a.end_time + 4 == b.start_time")

    def test_end_after_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.end_time.after(m.b.start_time, delay=-2))

        self.assertIsInstance(m.c.expr, BeforeExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertIs(m.c.expr.args[0], m.b.start_time)
        self.assertIs(m.c.expr.args[1], m.a.end_time)
        self.assertEqual(m.c.expr.delay, -2)

        self.assertEqual(str(m.c.expr), "b.start_time - 2 <= a.end_time")

    def test_end_before_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.end_time.before(m.b.end_time, delay=-5))

        self.assertIsInstance(m.c.expr, BeforeExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertIs(m.c.expr.args[0], m.a.end_time)
        self.assertIs(m.c.expr.args[1], m.b.end_time)
        self.assertEqual(m.c.expr.delay, -5)

        self.assertEqual(str(m.c.expr), "a.end_time - 5 <= b.end_time")

    def test_end_at_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.end_time.at(m.b.end_time, delay=-3))

        self.assertIsInstance(m.c.expr, AtExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertIs(m.c.expr.args[0], m.a.end_time)
        self.assertIs(m.c.expr.args[1], m.b.end_time)
        self.assertEqual(m.c.expr.delay, -3)

        self.assertEqual(str(m.c.expr), "a.end_time - 3 == b.end_time")

    def test_end_after_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.a.end_time.after(m.b.end_time))

        self.assertIsInstance(m.c.expr, BeforeExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertIs(m.c.expr.args[0], m.b.end_time)
        self.assertIs(m.c.expr.args[1], m.a.end_time)
        self.assertEqual(m.c.expr.delay, 0)

        self.assertEqual(str(m.c.expr), "b.end_time <= a.end_time")

    def test_end_before_start_param_delay(self):
        m = self.get_model()
        m.PrepTime = Param(initialize=5)
        m.c = LogicalConstraint(
            expr=m.a.end_time.before(m.b.start_time, delay=m.PrepTime)
        )
        self.assertIsInstance(m.c.expr, BeforeExpression)
        self.assertEqual(len(m.c.expr.args), 3)
        self.assertIs(m.c.expr.args[0], m.a.end_time)
        self.assertIs(m.c.expr.args[1], m.b.start_time)
        self.assertIs(m.c.expr.delay, m.PrepTime)

        self.assertEqual(str(m.c.expr), "a.end_time + PrepTime <= b.start_time")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Management of Pyomo commands
"""

import logging

logger = logging.getLogger('pyomo.common')


registry = {}


#
# Decorate functions that are Pyomo commands
#
def pyomo_command(name=None, doc=None):
    #
    def wrap(fn):
        if name is None:  # pragma:nocover
            logger.error("Error applying decorator.  No command name!")
            return
        if doc is None:  # pragma:nocover
            logger.error("Error applying decorator.  No command documentation!")
            return
        #
        global registry
        registry[name] = doc
        return fn

    #
    return wrap


def get_pyomo_commands():  # pragma:nocover
    return registry



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.alternative_solutions.aos_utils import logcontext
from pyomo.contrib.alternative_solutions.solution import Solution
from pyomo.contrib.alternative_solutions.solnpool import gurobi_generate_solutions
from pyomo.contrib.alternative_solutions.balas import enumerate_binary_solutions
from pyomo.contrib.alternative_solutions.obbt import (
    obbt_analysis,
    obbt_analysis_bounds_and_solutions,
)
from pyomo.contrib.alternative_solutions.lp_enum import enumerate_linear_solutions



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
import pyomo.environ as pyo
from pyomo.contrib import appsi
from pyomo.contrib.appsi.cmodel import cmodel_available
from pyomo.contrib.fbbt.tests.test_fbbt import FbbtTestBase
from pyomo.common.errors import InfeasibleConstraintException
import math

pe = pyo


@unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
class TestFbbt(FbbtTestBase, unittest.TestCase):
    def setUp(self) -> None:
        self.it = appsi.fbbt.IntervalTightener()
        self.tightener = self.it.perform_fbbt


@unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
class TestFBBTPersistent(unittest.TestCase):
    def test_persistent(self):
        m = pe.ConcreteModel()
        m.x = pe.Var()
        m.y = pe.Var()
        m.w = pe.Var()
        m.xl = pe.Param(mutable=True)
        m.xu = pe.Param(mutable=True)
        m.yl = pe.Param(mutable=True)
        m.yu = pe.Param(mutable=True)
        m.c1 = pe.Constraint(expr=m.w >= m.xl * m.y + m.x * m.yl - m.xl * m.yl)
        m.c2 = pe.Constraint(expr=m.w >= m.xu * m.y + m.x * m.yu - m.xu * m.yu)
        m.c3 = pe.Constraint(expr=m.w <= m.xu * m.y + m.x * m.yl - m.xu * m.yl)
        m.c4 = pe.Constraint(expr=m.w <= m.x * m.yu + m.xl * m.y - m.xl * m.yu)

        m.xl.value = -2
        m.xu.value = 2
        m.yl.value = -2
        m.yu.value = 2
        m.x.setlb(m.xl.value)
        m.x.setub(m.xu.value)
        m.y.setlb(m.yl.value)
        m.y.setub(m.yu.value)
        it = appsi.fbbt.IntervalTightener()
        it.perform_fbbt(m, symbolic_solver_labels=True)
        self.assertAlmostEqual(m.w.lb, -12)
        self.assertAlmostEqual(m.w.ub, 12)

        m.xl.value = -1
        m.xu.value = 1
        m.x.setlb(m.xl.value)
        m.x.setub(m.xu.value)
        it.perform_fbbt_with_seed(m, m.x)
        self.assertAlmostEqual(m.w.lb, -6)
        self.assertAlmostEqual(m.w.ub, 6)

        m.xl.value = -0.5
        m.xu.value = 0.5
        m.x.setlb(m.xl.value)
        m.x.setub(m.xu.value)
        it.perform_fbbt(m)
        self.assertAlmostEqual(m.w.lb, -3)
        self.assertAlmostEqual(m.w.ub, 3)

        m.yl.value = -1
        m.yu.value = 1
        m.y.setlb(m.yl.value)
        m.y.setub(m.yu.value)
        del m.c1
        del m.c2
        it.perform_fbbt(m)
        self.assertAlmostEqual(m.w.lb, -3)
        self.assertAlmostEqual(m.w.ub, 1.5)

    def test_sync_after_infeasible(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(bounds=(1, 1))
        m.y = pe.Var()
        m.c1 = pe.Constraint(expr=m.x == m.y)
        m.c2 = pe.Constraint(expr=m.y == 2)
        it = appsi.fbbt.IntervalTightener()
        try:
            it.perform_fbbt(m)
            was_infeasible = False
        except InfeasibleConstraintException:
            was_infeasible = True
        self.assertTrue(was_infeasible)
        self.assertAlmostEqual(m.x.lb, 1)
        self.assertAlmostEqual(m.x.ub, 1)
        self.assertAlmostEqual(m.y.lb, 1)
        self.assertAlmostEqual(m.y.ub, 1)

        m = pe.ConcreteModel()
        m.x = pe.Var(bounds=(1, 1))
        m.y = pe.Var()
        m.c1 = pe.Constraint(expr=m.x == m.y)
        m.c2 = pe.Constraint(expr=m.y == 2)
        it = appsi.fbbt.IntervalTightener()
        try:
            it.perform_fbbt_with_seed(m, m.x)
            was_infeasible = False
        except InfeasibleConstraintException:
            was_infeasible = True
        self.assertTrue(was_infeasible)
        self.assertAlmostEqual(m.x.lb, 1)
        self.assertAlmostEqual(m.x.ub, 1)
        self.assertAlmostEqual(m.y.lb, 1)
        self.assertAlmostEqual(m.y.ub, 1)

    def test_deactivated_constraints(self):
        m = pe.ConcreteModel()
        m.x = pe.Var()
        m.y = pe.Var()
        m.c1 = pe.Constraint(expr=m.x == 1)
        m.c2 = pe.Constraint(expr=m.y == m.x)
        it = appsi.fbbt.IntervalTightener()
        it.config.deactivate_satisfied_constraints = True
        it.perform_fbbt(m)
        self.assertFalse(m.c1.active)
        self.assertFalse(m.c2.active)
        m.c2.activate()
        m.x.setlb(0)
        m.x.setub(2)
        m.y.setlb(None)
        m.y.setub(None)
        it.perform_fbbt(m)
        self.assertTrue(m.c2.active)
        self.assertAlmostEqual(m.y.lb, 0)
        self.assertAlmostEqual(m.y.ub, 2)

    def test_named_exprs(self):
        m = pe.ConcreteModel()
        m.a = pe.Set(initialize=[1, 2, 3])
        m.x = pe.Var(m.a, bounds=(0, None))
        m.e = pe.Expression(m.a)
        for i in m.a:
            m.e[i].expr = i * m.x[i]
        m.c = pe.Constraint(expr=sum(m.e.values()) == 0)
        it = appsi.fbbt.IntervalTightener()
        it.perform_fbbt(m)
        for x in m.x.values():
            self.assertAlmostEqual(x.lb, 0)
            self.assertAlmostEqual(x.ub, 0)

    def test_named_exprs_nest(self):
        # test for issue #3184
        m = pe.ConcreteModel()
        m.x = pe.Var()
        m.e = pe.Expression(expr=m.x + 1)
        m.f = pe.Expression(expr=m.e)
        m.c = pe.Constraint(expr=(0, m.f, 0))
        it = appsi.fbbt.IntervalTightener()
        it.perform_fbbt(m)
        for x in m.x.values():
            self.assertAlmostEqual(x.lb, -1)
            self.assertAlmostEqual(x.ub, -1)



# -*- coding: utf-8 -*-
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import types
from itertools import islice

import logging
import traceback

logger = logging.getLogger('pyomo.core')

from pyomo.common.errors import PyomoException, DeveloperError
from pyomo.common.deprecation import (
    deprecation_warning,
    RenamedClass,
    relocated_module_attribute,
)
from .numvalue import (
    native_types,
    native_numeric_types,
    as_numeric,
    native_logical_types,
    value,
    is_potentially_variable,
)
from .base import ExpressionBase
from .boolean_value import BooleanValue, BooleanConstant
from .expr_common import _and, _or, _equiv, _inv, _xor, _impl, ExpressionType
from .numeric_expr import NumericExpression

import operator

relocated_module_attribute(
    'EqualityExpression',
    'pyomo.core.expr.relational_expr.EqualityExpression',
    version='6.4.3',
    f_globals=globals(),
)
relocated_module_attribute(
    'InequalityExpression',
    'pyomo.core.expr.relational_expr.InequalityExpression',
    version='6.4.3',
    f_globals=globals(),
)
relocated_module_attribute(
    'RangedExpression',
    'pyomo.core.expr.relational_expr.RangedExpression',
    version='6.4.3',
    f_globals=globals(),
)
relocated_module_attribute(
    'inequality',
    'pyomo.core.expr.relational_expr.inequality',
    version='6.4.3',
    f_globals=globals(),
)


def _generate_logical_proposition(etype, lhs, rhs):
    if (
        lhs.__class__ in native_types and lhs.__class__ not in native_logical_types
    ) and not isinstance(lhs, BooleanValue):
        return NotImplemented
    if (
        (rhs.__class__ in native_types and rhs.__class__ not in native_logical_types)
        and not isinstance(rhs, BooleanValue)
        and not (rhs is None and etype == _inv)
    ):
        return NotImplemented

    if etype == _equiv:
        return EquivalenceExpression((lhs, rhs))
    elif etype == _inv:
        assert rhs is None
        return NotExpression((lhs,))
    elif etype == _xor:
        return XorExpression((lhs, rhs))
    elif etype == _impl:
        return ImplicationExpression((lhs, rhs))
    elif etype == _and:
        return land(lhs, rhs)
    elif etype == _or:
        return lor(lhs, rhs)
    else:
        raise ValueError(
            "Unknown logical proposition type '%s'" % etype
        )  # pragma: no cover


class BooleanExpression(ExpressionBase, BooleanValue):
    """
    Logical expression base class.

    This class is used to define nodes in an expression
    tree.

    Abstract

    args:
        args (list or tuple): Children of this node.
    """

    __slots__ = ('_args_',)
    EXPRESSION_SYSTEM = ExpressionType.LOGICAL
    PRECEDENCE = 0

    def __init__(self, args):
        self._args_ = args

    @property
    def args(self):
        """
        Return the child nodes

        Returns: Either a list or tuple (depending on the node storage
            model) containing only the child nodes of this node
        """
        return self._args_[: self.nargs()]


class BooleanExpressionBase(metaclass=RenamedClass):
    __renamed__new_class__ = BooleanExpression
    __renamed__version__ = '6.4.3'


"""
---------------------------******************--------------------
The following methods are static methods for nodes creator. Those should
do the exact same thing as the class methods as well as overloaded operators.
"""


def lnot(Y):
    """
    Construct a NotExpression for the passed BooleanValue.
    """
    return NotExpression((Y,))


def equivalent(Y1, Y2):
    """
    Construct an EquivalenceExpression Y1 == Y2
    """
    return EquivalenceExpression((Y1, Y2))


def xor(Y1, Y2):
    """
    Construct an XorExpression Y1 xor Y2
    """
    return XorExpression((Y1, Y2))


def implies(Y1, Y2):
    """
    Construct an Implication using function, where Y1 implies Y2
    """
    return ImplicationExpression((Y1, Y2))


def _flattened(args):
    """Flatten any potentially indexed arguments."""
    for arg in args:
        if arg.__class__ in native_types:
            yield arg
        else:
            if isinstance(arg, (types.GeneratorType, list)):
                for _argdata in arg:
                    yield _argdata
            elif arg.is_indexed():
                for _argdata in arg.values():
                    yield _argdata
            else:
                yield arg


def _flattened_boolean_args(args):
    """Flatten any potentially indexed arguments and check that they are
    Boolean-valued."""
    for arg in args:
        if arg.__class__ in native_types:
            myiter = (arg,)
        elif isinstance(arg, (types.GeneratorType, list)):
            myiter = arg
        elif arg.is_indexed():
            myiter = arg.values()
        else:
            myiter = (arg,)
        for _argdata in myiter:
            if _argdata.__class__ in native_logical_types:
                yield _argdata
            elif hasattr(_argdata, 'is_logical_type') and _argdata.is_logical_type():
                yield _argdata
            elif isinstance(_argdata, BooleanValue):
                yield _argdata
            else:
                raise ValueError(
                    "Non-Boolean-valued argument '%s' encountered when constructing "
                    "expression of Boolean arguments" % arg
                )


def _flattened_numeric_args(args):
    """Flatten any potentially indexed arguments and check that they are
    numeric."""
    for arg in args:
        if arg.__class__ in native_types:
            myiter = (arg,)
        elif isinstance(arg, (types.GeneratorType, list)):
            myiter = arg
        elif arg.is_indexed():
            myiter = arg.values()
        else:
            myiter = (arg,)
        for _argdata in myiter:
            if _argdata.__class__ in native_numeric_types:
                yield _argdata
            elif hasattr(_argdata, 'is_numeric_type') and _argdata.is_numeric_type():
                yield _argdata
            else:
                raise ValueError(
                    "Non-numeric argument '%s' encountered when constructing "
                    "expression with numeric arguments" % arg
                )


def land(*args):
    """
    Construct an AndExpression between passed arguments.
    """
    result = AndExpression([])
    for argdata in _flattened_boolean_args(args):
        result = result.add(argdata)
    return result


def lor(*args):
    """
    Construct an OrExpression between passed arguments.
    """
    result = OrExpression([])
    for argdata in _flattened_boolean_args(args):
        result = result.add(argdata)
    return result


def exactly(n, *args):
    """Creates a new ExactlyExpression

    Require exactly n arguments to be True, to make the expression True

    Usage: exactly(2, m.Y1, m.Y2, m.Y3, ...)

    """
    result = ExactlyExpression([n] + list(_flattened_boolean_args(args)))
    return result


def atmost(n, *args):
    """Creates a new AtMostExpression

    Require at most n arguments to be True, to make the expression True

    Usage: atmost(2, m.Y1, m.Y2, m.Y3, ...)

    """
    result = AtMostExpression([n] + list(_flattened_boolean_args(args)))
    return result


def atleast(n, *args):
    """Creates a new AtLeastExpression

    Require at least n arguments to be True, to make the expression True

    Usage: atleast(2, m.Y1, m.Y2, m.Y3, ...)

    """
    result = AtLeastExpression([n] + list(_flattened_boolean_args(args)))
    return result


def all_different(*args):
    """Creates a new AllDifferentExpression

    Requires all of the arguments to take on a different value

    Usage: all_different(m.X1, m.X2, ...)
    """
    return AllDifferentExpression(list(_flattened_numeric_args(args)))


def count_if(*args):
    """Creates a new CountIfExpression

    Counts the number of True-valued arguments

    Usage: count_if(m.Y1, m.Y2, ...)
    """
    return CountIfExpression(list(_flattened_boolean_args(args)))


class UnaryBooleanExpression(BooleanExpression):
    """
    Abstract class for single-argument logical expressions.
    """

    def nargs(self):
        """
        Returns number of arguments in expression
        """
        return 1


class NotExpression(UnaryBooleanExpression):
    """
    This is the node for a NotExpression, this node should have exactly one child
    """

    PRECEDENCE = 2

    def getname(self, *arg, **kwd):
        return 'Logical Negation'

    def _to_string(self, values, verbose, smap):
        return "~%s" % values[0]

    def _apply_operation(self, result):
        return not result[0]


class BinaryBooleanExpression(BooleanExpression):
    """
    Abstract class for binary logical expressions.
    """

    def nargs(self):
        """
        Return the number of argument the expression has
        """
        return 2


class EquivalenceExpression(BinaryBooleanExpression):
    """
    Logical equivalence statement: Y_1 iff Y_2.

    """

    __slots__ = ()

    PRECEDENCE = 6

    def getname(self, *arg, **kwd):
        return 'iff'

    def _to_string(self, values, verbose, smap):
        return " iff ".join(values)

    def _apply_operation(self, result):
        return result[0] == result[1]


class XorExpression(BinaryBooleanExpression):
    """
    Logical Exclusive OR statement: Y_1 ⊻ Y_2
    """

    __slots__ = ()

    PRECEDENCE = 4

    def getname(self, *arg, **kwd):
        return 'xor'

    def _to_string(self, values, verbose, smap):
        return " ⊻ ".join(values)

    def _apply_operation(self, result):
        return operator.xor(result[0], result[1])


class ImplicationExpression(BinaryBooleanExpression):
    """
    Logical Implication statement: Y_1 --> Y_2.
    """

    __slots__ = ()

    PRECEDENCE = 6

    def getname(self, *arg, **kwd):
        return 'implies'

    def _to_string(self, values, verbose, smap):
        return " --> ".join(values)

    def _apply_operation(self, result):
        return (not result[0]) or result[1]


class NaryBooleanExpression(BooleanExpression):
    """
    The abstract class for NaryBooleanExpression.

    This class should never be initialized.
    """

    __slots__ = ('_nargs',)

    def __init__(self, args):
        self._args_ = args
        self._nargs = len(self._args_)

    def nargs(self):
        """
        Return the number of expression arguments
        """
        return self._nargs

    def getname(self, *arg, **kwd):
        return 'NaryBooleanExpression'


def _add_to_and_or_expression(orig_expr, new_arg):
    """
    Since AND and OR are Nary expressions, we extend the existing expression
    instead of creating a nested expression object if the types are compatible.
    """
    # Clone 'self', because AndExpression/OrExpression are immutable
    if new_arg.__class__ is orig_expr.__class__:
        # adding new AndExpression/OrExpression on the right
        new_expr = orig_expr.__class__(orig_expr._args_)
        new_expr._args_.extend(islice(new_arg._args_, new_arg._nargs))
    else:
        # adding new singleton on the right
        new_expr = orig_expr.__class__(orig_expr._args_)
        new_expr._args_.append(new_arg)

    # TODO set up id()-based scheme for avoiding duplicate entries

    new_expr._nargs = len(new_expr._args_)
    return new_expr


class AndExpression(NaryBooleanExpression):
    """
    This is the node for AndExpression.
    """

    __slots__ = ()

    PRECEDENCE = 3

    def getname(self, *arg, **kwd):
        return 'and'

    def _to_string(self, values, verbose, smap):
        return " ∧ ".join(values)

    def _apply_operation(self, result):
        return all(result)

    def add(self, new_arg):
        # FIXME: we should probably NOT perform this simplification
        # here, and instead use a dispatcher-based expression generation
        # system
        if new_arg.__class__ in native_logical_types:
            if new_arg:
                return self
            else:
                return BooleanConstant(False)
        return _add_to_and_or_expression(self, new_arg)


class OrExpression(NaryBooleanExpression):
    """
    This is the node for OrExpression.
    """

    __slots__ = ()

    PRECEDENCE = 5

    def getname(self, *arg, **kwd):
        return 'or'

    def _to_string(self, values, verbose, smap):
        return " ∨ ".join(values)

    def _apply_operation(self, result):
        return any(result)

    def add(self, new_arg):
        # FIXME: we should probably NOT perform this simplification
        # here, and instead use a dispatcher-based expression generation
        # system
        if new_arg.__class__ in native_logical_types:
            if new_arg:
                return BooleanConstant(True)
            else:
                return self
        return _add_to_and_or_expression(self, new_arg)


class ExactlyExpression(NaryBooleanExpression):
    """
    Logical constraint that exactly N child statements are True.

    The first argument N is expected to be a numeric non-negative integer.
    Subsequent arguments are expected to be Boolean.

    Usage: exactly(1, True, False, False) --> True

    """

    __slots__ = ()

    PRECEDENCE = 9

    def getname(self, *arg, **kwd):
        return 'exactly'

    def _to_string(self, values, verbose, smap):
        return "exactly(%s: [%s])" % (values[0], ", ".join(values[1:]))

    def _apply_operation(self, result):
        return sum(result[1:]) == result[0]


class AtMostExpression(NaryBooleanExpression):
    """
    Logical constraint that at most N child statements are True.

    The first argument N is expected to be a numeric non-negative integer.
    Subsequent arguments are expected to be Boolean.

    Usage: atmost(1, True, False, False) --> True

    """

    __slots__ = ()

    PRECEDENCE = 9

    def getname(self, *arg, **kwd):
        return 'atmost'

    def _to_string(self, values, verbose, smap):
        return "atmost(%s: [%s])" % (values[0], ", ".join(values[1:]))

    def _apply_operation(self, result):
        return sum(result[1:]) <= result[0]


class AtLeastExpression(NaryBooleanExpression):
    """
    Logical constraint that at least N child statements are True.

    The first argument N is expected to be a numeric non-negative integer.
    Subsequent arguments are expected to be Boolean.

    Usage: atleast(1, True, False, False) --> True

    """

    __slots__ = ()

    PRECEDENCE = 9

    def getname(self, *arg, **kwd):
        return 'atleast'

    def _to_string(self, values, verbose, smap):
        return "atleast(%s: [%s])" % (values[0], ", ".join(values[1:]))

    def _apply_operation(self, result):
        return sum(result[1:]) >= result[0]


class AllDifferentExpression(NaryBooleanExpression):
    """
    Logical expression that all of the N child statements have different values.
    All arguments are expected to be discrete-valued.
    """

    __slots__ = ()

    PRECEDENCE = None

    def getname(self, *arg, **kwd):
        return 'all_different'

    def _to_string(self, values, verbose, smap):
        return "all_different(%s)" % (", ".join(values))

    def _apply_operation(self, result):
        last = None
        # we know these are integer-valued, so we can just sort them an make
        # sure that no adjacent pairs have the same value.
        for val in sorted(result):
            if last == val:
                return False
            last = val
        return True


class CountIfExpression(NumericExpression):
    """
    Logical expression that returns the number of True child statements.
    All arguments are expected to be Boolean-valued.
    """

    __slots__ = ()
    PRECEDENCE = None

    # NumericExpression assumes binary operator, so we have to override.
    def nargs(self):
        return len(self._args_)

    def getname(self, *arg, **kwd):
        return 'count_if'

    def _to_string(self, values, verbose, smap):
        return "count_if(%s)" % (", ".join(values))

    def _apply_operation(self, result):
        return sum(value(r) for r in result)


special_boolean_atom_types = {ExactlyExpression, AtMostExpression, AtLeastExpression}



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import sys
from os import stat
from abc import ABC, abstractmethod
from io import StringIO

from pyomo.opt.base import ProblemFormat, ResultsFormat, OptSolver
from pyomo.opt.base.solvers import SolverFactory
from pyomo.common.collections import Bunch
from pyomo.common.dependencies import attempt_import
from pyomo.opt.results import (
    SolverResults,
    SolverStatus,
    TerminationCondition,
    SolutionStatus,
    ProblemSense,
)
from pyomo.common.tempfiles import TempfileManager
from pyomo.core.base import Var
from pyomo.core.base.block import BlockData
from pyomo.core.kernel.block import IBlock
from pyomo.common.log import LogStream
from pyomo.common.tee import capture_output, TeeStream

uuid, uuid_available = attempt_import('uuid')
logger = logging.getLogger("pyomo.solvers")


STATUS_TO_SOLVERSTATUS = {
    "OK": SolverStatus.ok,
    "SYNTAX_ERROR": SolverStatus.error,
    "DATA_ERROR": SolverStatus.error,
    "OUT_OF_MEMORY": SolverStatus.aborted,
    "IO_ERROR": SolverStatus.error,
    "ERROR": SolverStatus.error,
}

# This combines all status codes from OPTLP/solvelp and OPTMILP/solvemilp
SOLSTATUS_TO_TERMINATIONCOND = {
    "OPTIMAL": TerminationCondition.optimal,
    "OPTIMAL_AGAP": TerminationCondition.optimal,
    "OPTIMAL_RGAP": TerminationCondition.optimal,
    "OPTIMAL_COND": TerminationCondition.optimal,
    "TARGET": TerminationCondition.optimal,
    "CONDITIONAL_OPTIMAL": TerminationCondition.optimal,
    "FEASIBLE": TerminationCondition.feasible,
    "INFEASIBLE": TerminationCondition.infeasible,
    "UNBOUNDED": TerminationCondition.unbounded,
    "INFEASIBLE_OR_UNBOUNDED": TerminationCondition.infeasibleOrUnbounded,
    "SOLUTION_LIM": TerminationCondition.maxEvaluations,
    "NODE_LIM_SOL": TerminationCondition.maxEvaluations,
    "NODE_LIM_NOSOL": TerminationCondition.maxEvaluations,
    "ITERATION_LIMIT_REACHED": TerminationCondition.maxIterations,
    "TIME_LIM_SOL": TerminationCondition.maxTimeLimit,
    "TIME_LIM_NOSOL": TerminationCondition.maxTimeLimit,
    "TIME_LIMIT_REACHED": TerminationCondition.maxTimeLimit,
    "ABORTED": TerminationCondition.userInterrupt,
    "ABORT_SOL": TerminationCondition.userInterrupt,
    "ABORT_NOSOL": TerminationCondition.userInterrupt,
    "OUTMEM_SOL": TerminationCondition.solverFailure,
    "OUTMEM_NOSOL": TerminationCondition.solverFailure,
    "FAILED": TerminationCondition.solverFailure,
    "FAIL_SOL": TerminationCondition.solverFailure,
    "FAIL_NOSOL": TerminationCondition.solverFailure,
}


SOLSTATUS_TO_MESSAGE = {
    "OPTIMAL": "The solution is optimal.",
    "OPTIMAL_AGAP": "The solution is optimal within the absolute gap specified by the ABSOBJGAP= option.",
    "OPTIMAL_RGAP": "The solution is optimal within the relative gap specified by the RELOBJGAP= option.",
    "OPTIMAL_COND": "The solution is optimal, but some infeasibilities (primal, bound, or integer) exceed tolerances due to scaling or choice of a small INTTOL= value.",
    "TARGET": "The solution is not worse than the target specified by the TARGET= option.",
    "CONDITIONAL_OPTIMAL": "The solution is optimal, but some infeasibilities (primal, dual or bound) exceed tolerances due to scaling or preprocessing.",
    "FEASIBLE": "The problem is feasible. This status is displayed when the IIS=TRUE option is specified and the problem is feasible.",
    "INFEASIBLE": "The problem is infeasible.",
    "UNBOUNDED": "The problem is unbounded.",
    "INFEASIBLE_OR_UNBOUNDED": "The problem is infeasible or unbounded.",
    "SOLUTION_LIM": "The solver reached the maximum number of solutions specified by the MAXSOLS= option.",
    "NODE_LIM_SOL": "The solver reached the maximum number of nodes specified by the MAXNODES= option and found a solution.",
    "NODE_LIM_NOSOL": "The solver reached the maximum number of nodes specified by the MAXNODES= option and did not find a solution.",
    "ITERATION_LIMIT_REACHED": "The maximum allowable number of iterations was reached.",
    "TIME_LIM_SOL": "The solver reached the execution time limit specified by the MAXTIME= option and found a solution.",
    "TIME_LIM_NOSOL": "The solver reached the execution time limit specified by the MAXTIME= option and did not find a solution.",
    "TIME_LIMIT_REACHED": "The solver reached its execution time limit.",
    "ABORTED": "The solver was interrupted externally.",
    "ABORT_SOL": "The solver was stopped by the user but still found a solution.",
    "ABORT_NOSOL": "The solver was stopped by the user and did not find a solution.",
    "OUTMEM_SOL": "The solver ran out of memory but still found a solution.",
    "OUTMEM_NOSOL": "The solver ran out of memory and either did not find a solution or failed to output the solution due to insufficient memory.",
    "FAILED": "The solver failed to converge, possibly due to numerical issues.",
    "FAIL_SOL": "The solver stopped due to errors but still found a solution.",
    "FAIL_NOSOL": "The solver stopped due to errors and did not find a solution.",
}


@SolverFactory.register("sas", doc="The SAS LP/MIP solver")
class SAS(OptSolver):
    """The SAS optimization solver"""

    def __new__(cls, *args, **kwds):
        mode = kwds.pop("solver_io", None)
        if mode != None:
            return SolverFactory(mode, **kwds)
        else:
            # Choose solver factory automatically
            # based on what can be loaded.
            s = SolverFactory("_sas94", **kwds)
            if not s.available():
                s = SolverFactory("_sascas", **kwds)
            return s


class SASAbc(ABC, OptSolver):
    """Abstract base class for the SAS solver interfaces. Simply to avoid code duplication."""

    def __init__(self, **kwds):
        """Initialize the SAS solver interfaces."""
        kwds["type"] = "sas"
        super(SASAbc, self).__init__(**kwds)

        #
        # Set up valid problem formats and valid results for each
        # problem format
        #
        self._valid_problem_formats = [ProblemFormat.mps]
        self._valid_result_formats = {ProblemFormat.mps: [ResultsFormat.soln]}

        self._keepfiles = False
        self._capabilities.linear = True
        self._capabilities.integer = True

        super(SASAbc, self).set_problem_format(ProblemFormat.mps)

    def _presolve(self, *args, **kwds):
        """Set things up for the actual solve."""
        # create a context in the temporary file manager for
        # this plugin - is "pop"ed in the _postsolve method.
        TempfileManager.push()

        # Get the warmstart flag
        self.warmstart_flag = kwds.pop("warmstart", False)

        # Call parent presolve function
        super(SASAbc, self)._presolve(*args, **kwds)

        # Store the model, too bad this is not done in the base class
        for arg in args:
            if isinstance(arg, (BlockData, IBlock)):
                # Store the instance
                self._instance = arg
                self._vars = []
                for block in self._instance.block_data_objects(active=True):
                    for vardata in block.component_data_objects(
                        Var, active=True, descend_into=False
                    ):
                        self._vars.append(vardata)
                # Store the symbol map, we need this for example when writing the warmstart file
                if isinstance(self._instance, IBlock):
                    self._smap = getattr(self._instance, "._symbol_maps")[self._smap_id]
                else:
                    self._smap = self._instance.solutions.symbol_map[self._smap_id]

        # Create the primalin data
        if self.warmstart_flag:
            filename = self._warm_start_file_name = TempfileManager.create_tempfile(
                ".sol", text=True
            )
            smap = self._smap
            numWritten = 0
            with open(filename, "w") as file:
                file.write("_VAR_,_VALUE_\n")
                for var in self._vars:
                    if (var.value is not None) and (id(var) in smap.byObject):
                        name = smap.byObject[id(var)]
                        file.write(
                            "{name},{value}\n".format(name=name, value=var.value)
                        )
                        numWritten += 1
            if numWritten == 0:
                # No solution available, disable warmstart
                self.warmstart_flag = False

    def available(self, exception_flag=False):
        """True if the solver is available"""
        if not self._python_api_exists:
            return False
        return self.start_sas_session() is not None

    def _has_integer_variables(self):
        """True if the problem has integer variables."""
        for vardata in self._vars:
            if vardata.is_binary() or vardata.is_integer():
                return True
        return False

    def _create_results_from_status(self, status, solution_status):
        """Create a results object and set the status code and messages."""
        results = SolverResults()
        results.solver.name = "SAS"
        results.solver.status = STATUS_TO_SOLVERSTATUS[status]
        results.solver.hasSolution = False
        if results.solver.status == SolverStatus.ok:
            results.solver.termination_condition = SOLSTATUS_TO_TERMINATIONCOND[
                solution_status
            ]
            results.solver.message = results.solver.termination_message = (
                SOLSTATUS_TO_MESSAGE[solution_status]
            )
            results.solver.status = TerminationCondition.to_solver_status(
                results.solver.termination_condition
            )
            if "OPTIMAL" in solution_status or "_SOL" in solution_status:
                results.solver.hasSolution = True
        elif results.solver.status == SolverStatus.aborted:
            results.solver.termination_condition = TerminationCondition.userInterrupt
            if solution_status != "ERROR":
                results.solver.message = results.solver.termination_message = (
                    SOLSTATUS_TO_MESSAGE[solution_status]
                )
        else:
            results.solver.termination_condition = TerminationCondition.error
            results.solver.message = results.solver.termination_message = (
                SOLSTATUS_TO_MESSAGE["FAILED"]
            )
        return results

    @abstractmethod
    def _apply_solver(self):
        pass

    def _postsolve(self):
        """Clean up at the end, especially the temp files."""
        # Let the base class deal with returning results.
        results = super(SASAbc, self)._postsolve()

        # Finally, clean any temporary files registered with the temp file
        # manager, created populated *directly* by this plugin. does not
        # include, for example, the execution script. but does include
        # the warm-start file.
        TempfileManager.pop(remove=not self._keepfiles)

        return results

    def warm_start_capable(self):
        """True if the solver interface supports MILP warmstarting."""
        return True


@SolverFactory.register("_sas94", doc="SAS 9.4 interface")
class SAS94(SASAbc):
    """
    Solver interface for SAS 9.4 using saspy. See the saspy documentation about
    how to create a connection.
    The swat connection options can be specified on the SolverFactory call.
    """

    def __init__(self, **kwds):
        """Initialize the solver interface and see if the saspy package is available."""
        super(SAS94, self).__init__(**kwds)

        try:
            import saspy

            self._sas = saspy
        except ImportError:
            self._python_api_exists = False
        except Exception as e:
            self._python_api_exists = False
            # For other exceptions, raise it so that it does not get lost
            raise e
        else:
            self._python_api_exists = True
            self._sas.logger.setLevel(logger.level)

        # Store other options for the SAS session
        self._session_options = kwds
        self._sas_session = None

    def __del__(self):
        # Close the session, if we created one
        if self._sas_session:
            self._sas_session.endsas()
            del self._sas_session

    def _create_statement_str(self, statement):
        """Helper function to create the strings for the statements of the proc OPTLP/OPTMILP code."""
        stmt = self.options.pop(statement, None)
        if stmt:
            return (
                statement.strip()
                + " "
                + " ".join(option + "=" + str(value) for option, value in stmt.items())
                + ";"
            )
        else:
            return ""

    def sas_version(self):
        return self._sasver

    def start_sas_session(self):
        if self._sas_session is None:
            # Create (and cache) the session
            try:
                self._sas_session = self._sas.SASsession(**self._session_options)
            except:
                pass
        return self._sas_session

    def _apply_solver(self):
        """ "Prepare the options and run the solver. Then store the data to be returned."""
        logger.debug("Running SAS")

        # Set return code to issue an error if we get interrupted
        self._rc = -1

        # Figure out if the problem has integer variables
        with_opt = self.options.pop("with", None)
        if with_opt == "lp":
            proc = "OPTLP"
        elif with_opt == "milp":
            proc = "OPTMILP"
        else:
            # Check if there are integer variables, this might be slow
            proc = "OPTMILP" if self._has_integer_variables() else "OPTLP"

        # Get the rootnode options
        decomp_str = self._create_statement_str("decomp")
        decompmaster_str = self._create_statement_str("decompmaster")
        decompmasterip_str = self._create_statement_str("decompmasterip")
        decompsubprob_str = self._create_statement_str("decompsubprob")
        rootnode_str = self._create_statement_str("rootnode")

        # Get a unique identifier, always use the same with different prefixes
        unique = uuid.uuid4().hex[:16]

        # Create unique filename for output datasets
        primalout_dataset_name = "pout" + unique
        dualout_dataset_name = "dout" + unique
        primalin_dataset_name = None

        # Handle warmstart
        warmstart_str = ""
        if self.warmstart_flag:
            # Set the warmstart basis option
            primalin_dataset_name = "pin" + unique
            if proc != "OPTLP":
                warmstart_str = """
                                proc import datafile='{primalin}'
                                    out={primalin_dataset_name}
                                    dbms=csv
                                    replace;
                                    getnames=yes;
                                    run;
                                """.format(
                    primalin=self._warm_start_file_name,
                    primalin_dataset_name=primalin_dataset_name,
                )
                self.options["primalin"] = primalin_dataset_name

        # Convert options to string
        opt_str = " ".join(
            option + "=" + str(value) for option, value in self.options.items()
        )

        # Set some SAS options to make the log more clean
        sas_options = "option notes nonumber nodate nosource pagesize=max;"

        # Get the current SAS session, submit the code and return the results
        sas = self.start_sas_session()

        # Find the version of 9.4 we are using
        self._sasver = sas.sasver

        # Upload files, only if not accessible locally
        upload_mps = False
        if not sas.file_info(self._problem_files[0], quiet=True):
            sas.upload(self._problem_files[0], self._problem_files[0], overwrite=True)
            upload_mps = True

        upload_pin = False
        if self.warmstart_flag and not sas.file_info(
            self._warm_start_file_name, quiet=True
        ):
            sas.upload(
                self._warm_start_file_name, self._warm_start_file_name, overwrite=True
            )
            upload_pin = True

        # Using a function call to make it easier to mock the version check
        major_version = self.sas_version()[0]
        minor_version = self.sas_version().split("M", 1)[1][0]
        if major_version == "9" and int(minor_version) < 5:
            raise NotImplementedError(
                "Support for SAS 9.4 M4 and earlier is not implemented."
            )
        elif major_version == "9" and int(minor_version) == 5:
            # In 9.4M5 we have to create an MPS data set from an MPS file first
            # Earlier versions will not work because the MPS format in incompatible
            mps_dataset_name = "mps" + unique
            res = sas.submit(
                """
                            {sas_options}
                            {warmstart}
                            %MPS2SASD(MPSFILE="{mpsfile}", OUTDATA={mps_dataset_name}, MAXLEN=256, FORMAT=FREE);
                            proc {proc} data={mps_dataset_name} {options} primalout={primalout_dataset_name} dualout={dualout_dataset_name};
                            {decomp}
                            {decompmaster}
                            {decompmasterip}
                            {decompsubprob}
                            {rootnode}
                            run;
                            """.format(
                    sas_options=sas_options,
                    warmstart=warmstart_str,
                    proc=proc,
                    mpsfile=self._problem_files[0],
                    mps_dataset_name=mps_dataset_name,
                    options=opt_str,
                    primalout_dataset_name=primalout_dataset_name,
                    dualout_dataset_name=dualout_dataset_name,
                    decomp=decomp_str,
                    decompmaster=decompmaster_str,
                    decompmasterip=decompmasterip_str,
                    decompsubprob=decompsubprob_str,
                    rootnode=rootnode_str,
                ),
                results="TEXT",
            )
            sas.sasdata(mps_dataset_name).delete(quiet=True)
        else:
            # Since 9.4M6+ optlp/optmilp can read mps files directly (this includes Viya-based local installs)
            res = sas.submit(
                """
                            {sas_options}
                            {warmstart}
                            proc {proc} mpsfile=\"{mpsfile}\" {options} primalout={primalout_dataset_name} dualout={dualout_dataset_name};
                            {decomp}
                            {decompmaster}
                            {decompmasterip}
                            {decompsubprob}
                            {rootnode}
                            run;
                            """.format(
                    sas_options=sas_options,
                    warmstart=warmstart_str,
                    proc=proc,
                    mpsfile=self._problem_files[0],
                    options=opt_str,
                    primalout_dataset_name=primalout_dataset_name,
                    dualout_dataset_name=dualout_dataset_name,
                    decomp=decomp_str,
                    decompmaster=decompmaster_str,
                    decompmasterip=decompmasterip_str,
                    decompsubprob=decompsubprob_str,
                    rootnode=rootnode_str,
                ),
                results="TEXT",
            )

        # Delete uploaded file
        if upload_mps:
            sas.file_delete(self._problem_files[0], quiet=True)
        if self.warmstart_flag and upload_pin:
            sas.file_delete(self._warm_start_file_name, quiet=True)

        # Store log and ODS output
        self._log = res["LOG"]
        self._lst = res["LST"]
        if "ERROR 22-322: Syntax error" in self._log:
            raise ValueError(
                "An option passed to the SAS solver caused a syntax error: {log}".format(
                    log=self._log
                )
            )
        else:
            # Print log if requested by the user, only if we did not already print it
            if self._tee:
                print(self._log)
        self._macro = dict(
            (key.strip(), value.strip())
            for key, value in (
                pair.split("=") for pair in sas.symget("_OR" + proc + "_").split()
            )
        )
        if self._macro.get("STATUS", "ERROR") == "OK":
            primal_out = sas.sd2df(primalout_dataset_name)
            dual_out = sas.sd2df(dualout_dataset_name)

        # Delete data sets, they will go away automatically, but does not hurt to delete them
        if primalin_dataset_name:
            sas.sasdata(primalin_dataset_name).delete(quiet=True)
        sas.sasdata(primalout_dataset_name).delete(quiet=True)
        sas.sasdata(dualout_dataset_name).delete(quiet=True)

        # Prepare the solver results
        results = self.results = self._create_results_from_status(
            self._macro.get("STATUS", "ERROR"),
            self._macro.get("SOLUTION_STATUS", "ERROR"),
        )

        if "Objective Sense            Maximization" in self._lst:
            results.problem.sense = ProblemSense.maximize
        else:
            results.problem.sense = ProblemSense.minimize

        # Prepare the solution information
        if results.solver.hasSolution:
            sol = results.solution.add()

            # Store status in solution
            sol.status = SolutionStatus.feasible
            sol.termination_condition = SOLSTATUS_TO_TERMINATIONCOND[
                self._macro.get("SOLUTION_STATUS", "ERROR")
            ]

            # Store objective value in solution
            sol.objective["__default_objective__"] = {"Value": self._macro["OBJECTIVE"]}

            if proc == "OPTLP":
                # Convert primal out data set to variable dictionary
                # Use pandas functions for efficiency
                primal_out = primal_out[["_VAR_", "_VALUE_", "_STATUS_", "_R_COST_"]]
                primal_out = primal_out.set_index("_VAR_", drop=True)
                primal_out = primal_out.rename(
                    {"_VALUE_": "Value", "_STATUS_": "Status", "_R_COST_": "rc"},
                    axis="columns",
                )
                sol.variable = primal_out.to_dict("index")

                # Convert dual out data set to constraint dictionary
                # Use pandas functions for efficiency
                dual_out = dual_out[["_ROW_", "_VALUE_", "_STATUS_", "_ACTIVITY_"]]
                dual_out = dual_out.set_index("_ROW_", drop=True)
                dual_out = dual_out.rename(
                    {"_VALUE_": "dual", "_STATUS_": "Status", "_ACTIVITY_": "slack"},
                    axis="columns",
                )
                sol.constraint = dual_out.to_dict("index")
            else:
                # Convert primal out data set to variable dictionary
                # Use pandas functions for efficiency
                primal_out = primal_out[["_VAR_", "_VALUE_"]]
                primal_out = primal_out.set_index("_VAR_", drop=True)
                primal_out = primal_out.rename({"_VALUE_": "Value"}, axis="columns")
                sol.variable = primal_out.to_dict("index")

        self._rc = 0
        return Bunch(rc=self._rc, log=self._log)


@SolverFactory.register("_sascas", doc="SAS Viya CAS Server interface")
class SASCAS(SASAbc):
    """
    Solver interface connection to a SAS Viya CAS server using swat.
    See the documentation for the swat package about how to create a connection.
    The swat connection options can be specified on the SolverFactory call.
    """

    def __init__(self, **kwds):
        """Initialize and try to load the swat package."""
        super(SASCAS, self).__init__(**kwds)

        try:
            import swat

            self._sas = swat
        except ImportError:
            self._python_api_exists = False
        except Exception as e:
            self._python_api_exists = False
            # For other exceptions, raise it so that it does not get lost
            raise e
        else:
            self._python_api_exists = True

        self._session_options = kwds
        self._sas_session = None

    def __del__(self):
        # Close the session, if we created one
        if self._sas_session:
            self._sas_session.close()
            del self._sas_session

    def start_sas_session(self):
        if self._sas_session is None:
            # Create (and cache) the session
            try:
                self._sas_session = self._sas.CAS(**self._session_options)
            except:
                pass
        return self._sas_session

    def _uploadMpsFile(self, s, unique):
        # Declare a unique table name for the mps table
        mpsdata_table_name = "mps" + unique

        # Upload mps file to CAS, if the file is larger than 2 GB, we need to use convertMps instead of loadMps
        # Note that technically it is 2 Gibibytes file size that trigger the issue, but 2 GB is the safer threshold
        if stat(self._problem_files[0]).st_size > 2e9:
            # For files larger than 2 GB (this is a limitation of the loadMps action used in the else part).
            # Use convertMPS, first create file for upload.
            mpsWithIdFileName = TempfileManager.create_tempfile(".mps.csv", text=True)
            with open(mpsWithIdFileName, "w") as mpsWithId:
                mpsWithId.write("_ID_\tText\n")
                with open(self._problem_files[0], "r") as f:
                    id = 0
                    for line in f:
                        id += 1
                        mpsWithId.write(str(id) + "\t" + line.rstrip() + "\n")

            # Upload .mps.csv file
            mpscsv_table_name = "csv" + unique
            s.upload_file(
                mpsWithIdFileName,
                casout={"name": mpscsv_table_name, "replace": True},
                importoptions={"filetype": "CSV", "delimiter": "\t"},
            )

            # Convert .mps.csv file to .mps
            s.optimization.convertMps(
                data=mpscsv_table_name,
                casOut={"name": mpsdata_table_name, "replace": True},
                format="FREE",
                maxLength=256,
            )

            # Delete the table we don't need anymore
            if mpscsv_table_name:
                s.dropTable(name=mpscsv_table_name, quiet=True)
        else:
            # For small files (less than 2 GB), use loadMps
            with open(self._problem_files[0], "r") as mps_file:
                s.optimization.loadMps(
                    mpsFileString=mps_file.read(),
                    casout={"name": mpsdata_table_name, "replace": True},
                    format="FREE",
                    maxLength=256,
                )
        return mpsdata_table_name

    def _uploadPrimalin(self, s, unique):
        # Upload warmstart file to CAS with a unique name
        primalin_table_name = "pin" + unique
        s.upload_file(
            self._warm_start_file_name,
            casout={"name": primalin_table_name, "replace": True},
            importoptions={"filetype": "CSV"},
        )
        self.options["primalin"] = primalin_table_name
        return primalin_table_name

    def _retrieveSolution(
        self, s, r, results, action, primalout_table_name, dualout_table_name
    ):
        # Create solution
        sol = results.solution.add()

        # Store status in solution
        sol.status = SolutionStatus.feasible
        sol.termination_condition = SOLSTATUS_TO_TERMINATIONCOND[
            r.get("solutionStatus", "ERROR")
        ]

        # Store objective value in solution
        sol.objective["__default_objective__"] = {"Value": r["objective"]}

        if action == "solveMilp":
            primal_out = s.CASTable(name=primalout_table_name)
            # Use pandas functions for efficiency
            primal_out = primal_out[["_VAR_", "_VALUE_"]]
            sol.variable = {}
            for row in primal_out.itertuples(index=False):
                sol.variable[row[0]] = {"Value": row[1]}
        else:
            # Convert primal out data set to variable dictionary
            # Use panda functions for efficiency
            primal_out = s.CASTable(name=primalout_table_name)
            primal_out = primal_out[["_VAR_", "_VALUE_", "_STATUS_", "_R_COST_"]]
            sol.variable = {}
            for row in primal_out.itertuples(index=False):
                sol.variable[row[0]] = {"Value": row[1], "Status": row[2], "rc": row[3]}

            # Convert dual out data set to constraint dictionary
            # Use pandas functions for efficiency
            dual_out = s.CASTable(name=dualout_table_name)
            dual_out = dual_out[["_ROW_", "_VALUE_", "_STATUS_", "_ACTIVITY_"]]
            sol.constraint = {}
            for row in dual_out.itertuples(index=False):
                sol.constraint[row[0]] = {
                    "dual": row[1],
                    "Status": row[2],
                    "slack": row[3],
                }

    def _apply_solver(self):
        """ "Prepare the options and run the solver. Then store the data to be returned."""
        logger.debug("Running SAS Viya")

        # Set return code to issue an error if we get interrupted
        self._rc = -1

        # Figure out if the problem has integer variables
        with_opt = self.options.pop("with", None)
        if with_opt == "lp":
            action = "solveLp"
        elif with_opt == "milp":
            action = "solveMilp"
        else:
            # Check if there are integer variables, this might be slow
            action = "solveMilp" if self._has_integer_variables() else "solveLp"

        # Get a unique identifier, always use the same with different prefixes
        unique = uuid.uuid4().hex[:16]

        # Creat the output stream, we want to print to a log string as well as to the console
        self._log = StringIO()
        ostreams = [LogStream(level=logging.INFO, logger=logger)]
        ostreams.append(self._log)
        if self._tee:
            ostreams.append(sys.stdout)

        # Connect to CAS server
        with capture_output(output=TeeStream(*ostreams), capture_fd=False):
            s = self.start_sas_session()
            try:
                # Load the optimization action set
                s.loadactionset("optimization")

                mpsdata_table_name = self._uploadMpsFile(s, unique)

                primalin_table_name = None
                if self.warmstart_flag:
                    primalin_table_name = self._uploadPrimalin(s, unique)

                # Define output table names
                primalout_table_name = "pout" + unique
                dualout_table_name = None

                # Solve the problem in CAS
                if action == "solveMilp":
                    r = s.optimization.solveMilp(
                        data={"name": mpsdata_table_name},
                        primalOut={"name": primalout_table_name, "replace": True},
                        **self.options,
                    )
                else:
                    dualout_table_name = "dout" + unique
                    r = s.optimization.solveLp(
                        data={"name": mpsdata_table_name},
                        primalOut={"name": primalout_table_name, "replace": True},
                        dualOut={"name": dualout_table_name, "replace": True},
                        **self.options,
                    )

                # Prepare the solver results
                if r:
                    # Get back the primal and dual solution data sets
                    results = self.results = self._create_results_from_status(
                        r.get("status", "ERROR"), r.get("solutionStatus", "ERROR")
                    )

                    if results.solver.status != SolverStatus.error:
                        if r.ProblemSummary["cValue1"][1] == "Maximization":
                            results.problem.sense = ProblemSense.maximize
                        else:
                            results.problem.sense = ProblemSense.minimize

                        # Prepare the solution information
                        if results.solver.hasSolution:
                            self._retrieveSolution(
                                s,
                                r,
                                results,
                                action,
                                primalout_table_name,
                                dualout_table_name,
                            )
                    else:
                        raise ValueError("The SAS solver returned an error status.")
                else:
                    results = self.results = SolverResults()
                    results.solver.name = "SAS"
                    results.solver.status = SolverStatus.error
                    raise ValueError(
                        "An option passed to the SAS solver caused a syntax error."
                    )

            finally:
                if mpsdata_table_name:
                    s.dropTable(name=mpsdata_table_name, quiet=True)
                if primalin_table_name:
                    s.dropTable(name=primalin_table_name, quiet=True)
                if primalout_table_name:
                    s.dropTable(name=primalout_table_name, quiet=True)
                if dualout_table_name:
                    s.dropTable(name=dualout_table_name, quiet=True)

        self._log = self._log.getvalue()
        self._rc = 0
        return Bunch(rc=self._rc, log=self._log)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
#
"""Pyomo Units Checking Module
This module has some helpful methods to support checking units on Pyomo
module objects.
"""

import logging

from pyomo.core.base.units_container import units, UnitsError
from pyomo.core.base import (
    Objective,
    Constraint,
    Var,
    Param,
    Suffix,
    Set,
    SetOf,
    RangeSet,
    Block,
    ExternalFunction,
    Expression,
    value,
    BooleanVar,
    BuildAction,
    BuildCheck,
)
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.network import Port, Arc
from pyomo.mpec import Complementarity
from pyomo.gdp import Disjunct, Disjunction
from pyomo.core.expr.template_expr import IndexTemplate
from pyomo.core.expr.numvalue import native_types
from pyomo.util.components import iter_component
from pyomo.common.collections import ComponentSet

logger = logging.getLogger(__name__)


def check_units_equivalent(*args):
    """
    Returns True if the units associated with each of the
    expressions passed as arguments are all equivalent (and False
    otherwise).

    Note that this method will raise an exception if the units are
    inconsistent within an expression (since the units for that
    expression are not valid).

    Parameters
    ----------
    args : an argument list of Pyomo expressions

    Returns
    -------
    bool : True if all the expressions passed as arguments have the same units
    """
    try:
        assert_units_equivalent(*args)
        return True
    except UnitsError:
        return False


def assert_units_equivalent(*args):
    """
    Raise an exception if the units are inconsistent within an
    expression, or not equivalent across all the passed
    expressions.

    Parameters
    ----------
    args : an argument list of Pyomo expressions
        The Pyomo expressions to test

    Raises
    ------
    :py:class:`pyomo.core.base.units_container.UnitsError`, :py:class:`pyomo.core.base.units_container.InconsistentUnitsError`
    """
    # this call will raise an exception if an inconsistency is found
    pint_units = [units._get_pint_units(arg) for arg in args]
    pint_unit_compare = pint_units[0]
    for pint_unit in pint_units:
        if not units._equivalent_pint_units(pint_unit_compare, pint_unit):
            raise UnitsError(
                "Units between {} and {} are not consistent.".format(
                    str(pint_unit_compare), str(pint_unit)
                )
            )


def _assert_units_consistent_constraint_data(condata):
    """
    Raise an exception if the any units in lower, body, upper on a
    ConstraintData object are not consistent or are not equivalent
    with each other.
    """
    # Pyomo can rearrange expressions, resulting in a value
    # of 0 for upper or lower that does not have units associated
    # Therefore, if the lower and/or upper is 0, we allow it to be unitless
    # and check the consistency of the body only
    args = list()
    if condata.lower is not None and value(condata.lower) != 0.0:
        args.append(condata.lower)

    args.append(condata.body)

    if condata.upper is not None and value(condata.upper) != 0.0:
        args.append(condata.upper)

    if len(args) == 1:
        assert_units_consistent(*args)
    else:
        assert_units_equivalent(*args)


def _assert_units_consistent_arc_data(arcdata):
    """
    Raise an exception if the any units do not match for the connected ports
    """
    sport = arcdata.source
    dport = arcdata.destination
    if sport is None or dport is None:
        # nothing to check
        return

    # both sport and dport are not None
    # iterate over the vars in one and check against the other
    for key in sport.vars:
        svar = sport.vars[key]
        dvar = dport.vars[key]

        if svar.is_indexed():
            for k in svar:
                svardata = svar[k]
                dvardata = dvar[k]
                assert_units_equivalent(svardata, dvardata)
        else:
            assert_units_equivalent(svar, dvar)


def _assert_units_consistent_property_expr(obj):
    """
    Check the .expr property of the object and raise
    an exception if the units are not consistent
    """
    _assert_units_consistent_expression(obj.expr)


def _assert_units_consistent_expression(expr):
    """
    Raise an exception if any units in expr are inconsistent.
    """
    # this will raise an exception if the units are not consistent
    # in the expression
    pint_unit = units._get_pint_units(expr)
    # pyomo_unit = units.get_units(expr)


# Complementarities that are not in standard form do not
# current work with the checking code. The Units container
# should be modified to allow sum and relationals with zero
# terms (e.g., unitless). Then this code can be enabled.
# def _assert_units_complementarity(cdata):
#    """
#    Raise an exception if any units in either of the complementarity
#    expressions are inconsistent, and also check the standard block
#    methods.
#    """
#    if cdata._args[0] is not None:
#        pyomo_unit, pint_unit = units._get_units_tuple(cdata._args[0])
#    if cdata._args[1] is not None:
#        pyomo_unit, pint_unit = units._get_units_tuple(cdata._args[1])
#    _assert_units_consistent_block(cdata)


def _assert_units_consistent_block(obj):
    """
    This method gets all the components from the block
    and checks if the units are consistent on each of them
    """
    # check all the component objects
    for component in obj.component_objects(descend_into=False, active=True):
        assert_units_consistent(component)


_component_data_handlers = {
    Objective: _assert_units_consistent_property_expr,
    Constraint: _assert_units_consistent_constraint_data,
    Var: _assert_units_consistent_expression,
    DerivativeVar: _assert_units_consistent_expression,
    Port: None,
    Arc: _assert_units_consistent_arc_data,
    Expression: _assert_units_consistent_property_expr,
    Suffix: None,
    Param: _assert_units_consistent_expression,
    Disjunct: _assert_units_consistent_block,
    Disjunction: None,
    BooleanVar: None,
    Block: _assert_units_consistent_block,
    ExternalFunction: None,
    # TODO: change this when Sets / ContinuousSets sets have units:
    ContinuousSet: None,
    Set: None,
    SetOf: None,
    RangeSet: None,
    # TODO: Piecewise: _assert_units_consistent_piecewise,
    # TODO: SOSConstraint: _assert_units_consistent_sos,
    # TODO: LogicalConstraint: _assert_units_consistent_logical,
    BuildAction: None,
    BuildCheck: None,
    # complementarities that are not in normal form are not working yet
    # see comment in test_check_units
    # TODO: Complementarity: _assert_units_complementarity
}


def assert_units_consistent(obj):
    """
    This method raises an exception if the units are not
    consistent on the passed in object.  Argument obj can be one
    of the following components: Pyomo Block (or Model),
    Constraint, Objective, Expression, or it can be a Pyomo
    expression object

    Parameters
    ----------
    obj : Pyomo component (e.g., Block, Model, Constraint, Objective, or Expression) or Pyomo expression
       The object or expression to test

    Raises
    ------
    :py:class:`pyomo.core.base.units_container.UnitsError`, :py:class:`pyomo.core.base.units_container.InconsistentUnitsError`
    """
    objtype = type(obj)
    if objtype in native_types:
        return
    elif obj.is_expression_type() or objtype is IndexTemplate:
        try:
            _assert_units_consistent_expression(obj)
        except UnitsError:
            logger.error('Units problem with expression {}'.format(obj))
            raise
        return

    # if object is not in our component handler, raise an exception
    if obj.ctype not in _component_data_handlers:
        raise TypeError(
            "Units checking not supported for object of type {}.".format(obj.ctype)
        )

    # get the function form the list of handlers
    handler = _component_data_handlers[obj.ctype]
    if handler is None:
        return

    if obj.is_indexed():
        # check all the component data objects
        for cdata in obj.values():
            try:
                handler(cdata)
            except UnitsError:
                logger.error('Error in units when checking {}'.format(cdata))
                raise
    else:
        handler(obj)


def identify_inconsistent_units(block):
    """
    This function generates a ComponentSet of all Constraints, Expressions, and Objectives
    in a Block or model which have inconsistent units.

    Parameters
    ----------
    block : Pyomo Block or Model to test

    Returns
    ------
    ComponentSet : contains all Constraints, Expressions or Objectives which were
        identified as having unit consistency issues
    """
    # It would be nice (and more efficient) if there were a method that would check
    # unit consistency and return a bool for success or failure.
    # However, the underlying methods (at least as deep as I looked) all raise exceptions
    # so we need to iterate over the block here and do a try/except for each component

    inconsistent_units = ComponentSet()
    for obj in block.component_data_objects(
        [Constraint, Expression, Objective], descend_into=True
    ):
        try:
            assert_units_consistent(obj)
        except UnitsError:
            inconsistent_units.add(obj)
    return inconsistent_units



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import io
from typing import Sequence, Optional, Mapping

from pyomo.common.collections import ComponentMap
from pyomo.common.errors import MouseTrap
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.var import VarData
from pyomo.core.expr import value
from pyomo.core.staleflag import StaleFlagManager
from pyomo.core.expr.visitor import replace_expressions
from pyomo.repn.plugins.nl_writer import NLWriterInfo

from pyomo.contrib.solver.common.util import SolverError
from pyomo.contrib.solver.common.results import (
    Results,
    SolutionStatus,
    TerminationCondition,
)
from pyomo.contrib.solver.common.solution_loader import SolutionLoaderBase


class ASLSolFileData:
    """
    Defines the data fields found within an ASL .sol file
    """

    def __init__(self) -> None:
        self.message: str = None
        self.objno: int = 0
        self.solve_code: int = None
        self.ampl_options: list[int | float] = None
        self.primals: list[float] = None
        self.duals: list[float] = None
        self.var_suffixes: dict[str, dict[int, int | float]] = {}
        self.con_suffixes: dict[str, dict[int, int | float]] = {}
        self.obj_suffixes: dict[str, dict[int, int | float]] = {}
        self.problem_suffixes: dict[str, int | float] = {}
        self.suffix_table: dict[(int, str), list[int | float, str, ...]] = {}
        self.unparsed: str = None


class ASLSolFileSolutionLoader(SolutionLoaderBase):
    """
    Loader for solvers that create ASL .sol files (e.g., ipopt)
    """

    def __init__(self, sol_data: ASLSolFileData, nl_info: NLWriterInfo) -> None:
        self._sol_data = sol_data
        self._nl_info = nl_info

    def load_vars(self, vars_to_load: Optional[Sequence[VarData]] = None) -> None:
        if vars_to_load is not None:
            # If we are given a list of variables to load, it is easiest
            # to use the filtering in get_primals and then just set
            # those values.
            for var, val in self.get_primals(vars_to_load).items():
                var.set_value(val, skip_validation=True)
            StaleFlagManager.mark_all_as_stale(delayed=True)
            return

        if not self._sol_data.primals:
            # SOL file contained no primal values
            assert len(self._nl_info.variables) == 0
        else:
            # Load the primals provided by the SOL file (scaling if necessary)
            assert len(self._nl_info.variables) == len(self._sol_data.primals)
            if self._nl_info.scaling:
                for var, val, scale in zip(
                    self._nl_info.variables,
                    self._sol_data.primals,
                    self._nl_info.scaling.variables,
                ):
                    var.set_value(val / scale, skip_validation=True)
            else:
                for var, val in zip(self._nl_info.variables, self._sol_data.primals):
                    var.set_value(val, skip_validation=True)

        # Compute all variables presolved out of the model
        for var, v_expr in self._nl_info.eliminated_vars:
            var.set_value(value(v_expr), skip_validation=True)

        StaleFlagManager.mark_all_as_stale(delayed=True)

    def get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        result = ComponentMap()
        if not self._sol_data.primals:
            # SOL file contained no primal values
            assert len(self._nl_info.variables) == 0
        else:
            # Load the primals provided by the SOL file (scaling if necessary)
            assert len(self._nl_info.variables) == len(self._sol_data.primals)
            if self._nl_info.scaling:
                for var, val, scale in zip(
                    self._nl_info.variables,
                    self._sol_data.primals,
                    self._nl_info.scaling.variables,
                ):
                    result[var] = val / scale
            else:
                for var, val in zip(self._nl_info.variables, self._sol_data.primals):
                    result[var] = val

        # If we have eliminated variables, then we need to compute
        # them.  Unfortunately, the expressions that we kept are in
        # terms of the actual variable values (which we don't want to
        # modify).  We will make use of an expression replacement
        # visitor to perform the substitution and computation.
        #
        # It would be great if we could do this without creating the
        # entire (unfiltered) result, but we just don't (easily) know
        # which variable values we are going to need (either in the
        # vars_to_load list, or in any expression that might be needed
        # to compute an eliminated variable value.  So to keep things
        # simple (i.e., fewer bugs), we will go ahead and always compute
        # everything.
        if self._nl_info.eliminated_vars:
            val_map = {id(k): v for k, v in result.items()}
            for var, v_expr in self._nl_info.eliminated_vars:
                val = value(replace_expressions(v_expr, substitution_map=val_map))
                val_map[id(var)] = val
                result[var] = val

        if vars_to_load is not None:
            result = ComponentMap((v, result[v]) for v in vars_to_load)

        return result

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> dict[ConstraintData, float]:
        if len(self._nl_info.eliminated_vars) > 0:
            raise MouseTrap(
                'Complete duals are not available when variables have '
                'been presolved from the model.  Turn presolve off '
                '(solver.config.writer_config.linear_presolve=False) to get '
                'dual variable values.'
            )

        scaling = self._nl_info.scaling
        if scaling:
            _iter = zip(
                self._nl_info.constraints, self._sol_data.duals, scaling.constraints
            )
            inv_obj_scale = 1.0
            if self._nl_info.scaling.objectives:
                inv_obj_scale /= self._nl_info.scaling.objectives[self._sol_data.objno]
        else:
            _iter = zip(self._nl_info.constraints, self._sol_data.duals)
        if cons_to_load is not None:
            cons_to_load = set(cons_to_load)
            _iter = filter(lambda x: x[0] in cons_to_load, _iter)
        if scaling:
            return {con: val * scale * inv_obj_scale for con, val, scale in _iter}
        else:
            return {con: val for con, val in _iter}


def asl_solve_code_to_solution_status(
    sol_data: ASLSolFileData, result: Results
) -> None:
    """Convert the ASL "solve code" integer into a Pyomo status

    The ASL returns an indication of the solution status and termination
    condition using a single "solve code" integer.  This function
    implements the translation of the numeric value into the Pyomo
    equivalents (:class:`TerminationCondition` and
    :class:`SolutionStatus`), as well as a general string description,
    using the table from Section 14.2 in the AMPL Book [FGK02]_.

    """
    #
    # This table (the values and the string interpretations) are from
    # Chapter 14 in the AMPL Book:
    #
    code = sol_data.solve_code
    status = SolutionStatus.unknown if sol_data.primals else SolutionStatus.noSolution
    if code is None:
        message = f"AMPL({code}): solver did not generate a SOL file"
        term = TerminationCondition.error
    elif (code >= 0) and (code <= 99):
        # message = f"AMPL({code}:solved): optimal solution found"
        message = ''
        status = SolutionStatus.optimal
        term = TerminationCondition.convergenceCriteriaSatisfied
    elif (code >= 100) and (code <= 199):
        message = f"AMPL({code}:solved?): optimal solution indicated, but error likely"
        status = SolutionStatus.feasible
        term = TerminationCondition.error
    elif (code >= 200) and (code <= 299):
        message = f"AMPL({code}:infeasible): constraints cannot be satisfied"
        status = SolutionStatus.infeasible
        term = TerminationCondition.locallyInfeasible
    elif (code >= 300) and (code <= 399):
        message = f"AMPL({code}:unbounded): objective can be improved without limit"
        term = TerminationCondition.unbounded
    elif (code >= 400) and (code <= 499):
        message = f"AMPL({code}:limit): stopped by a limit that you set"
        term = TerminationCondition.iterationLimit  # this is not always correct
    elif (code >= 500) and (code <= 599):
        message = f"AMPL({code}:failure): stopped by an error condition in the solver"
        term = TerminationCondition.error
    else:
        message = f"AMPL({code}): unexpected solve code"
        term = TerminationCondition.error

    if sol_data.message:
        # TBD: [JDS 10/2025]: Why do we convert newlines to semicolons?
        result.extra_info.solver_message = sol_data.message.replace('\n', '; ')
        if message:
            result.extra_info.solver_message += '; ' + message
    else:
        result.extra_info.solver_message = message
    result.solution_status = status
    result.termination_condition = term


def parse_asl_sol_file(FILE: io.TextIOBase) -> ASLSolFileData:
    """Parse an ASL .sol file.

    This is a standalone routine to parse the AMPL Solver Library (ASL)
    "``.sol``" file format.  The resulting :class:`ASLSolFileData`
    object is a faithful representation of the data from the file.
    Translating the parsed data back into the context of a Pyomo model
    requires additional information (at the very least, the Pyomo model
    and the :class:`NLWriterInfo` data structure generated by the writer
    that originally created the ``.nl`` file that was sent to the
    solver.

    """
    sol_data = ASLSolFileData()

    # Parse the initial solver message and the AMPL options sections
    z = _parse_message_and_options(FILE, sol_data)

    #
    # Parse the duals and variable values
    #
    num_duals = z[1]  # "m" in writesol.c
    assert num_duals == z[0] or not num_duals
    sol_data.duals = [float(FILE.readline()) for i in range(num_duals)]

    num_primals = z[3]  # "n" in writesol.c
    assert num_primals == z[2] or not num_primals
    sol_data.primals = [float(FILE.readline()) for i in range(num_primals)]

    # Parse the OBJNO (objective number and solver exit code)
    _parse_objno_and_exitcode(FILE, sol_data)

    # Parse the suffix data
    _parse_suffixes(FILE, sol_data)

    return sol_data


def _parse_message_and_options(FILE: io.TextIOBase, data: ASLSolFileData) -> list[int]:
    msg = []
    # Some solvers (minto) do not write a message.  We will assume
    # all non-blank lines up the 'Options' line is the message.
    while True:
        line = FILE.readline()
        if not line:
            # EOF
            raise SolverError("Error reading `sol` file: no 'Options' line found.")
        line = line.strip()
        if line == 'Options':
            break
        if line:
            msg.append(line)
    data.message = "\n".join(msg)

    # WARNING: This appears to be undocumented outside of the ASL
    # writesol.c implementation.  Before changing this logic, please
    # familiarize yourself with that code.
    #
    # The AMPL options are a sequence of ints, the first of which
    # specifies the number of options to expect, followed by the
    # options (all ints), followed by the 4 int-elements of "z".
    #
    n_opts = int(FILE.readline())
    #
    # The ASL will occasionally "lie" about the number of options: if
    # the second option (not including the number of options) is "3",
    # then the ASL will add 2 to the number of options reported, and
    # will add *one* option (vbtol, a float) *after* the elements of
    # "z".
    #
    # Because of this, we will read the first two options from the file
    # first so we can know how to correctly parse the remaining options.
    assert n_opts >= 2
    ampl_options = [int(FILE.readline()), int(FILE.readline())]
    read_vbtol = ampl_options[1] == 3
    if read_vbtol:
        n_opts -= 2
    ampl_options.extend(int(FILE.readline()) for i in range(n_opts - 2))
    # Note: "z" comes from the name used for this data structure in
    # `writesol.c`.  It is unknown to us what motivated that name.
    #
    # Z: [ #cons; #duals, #vars, #var_vals ]
    #    #duals will either be #cons or 0
    #    #var_vals will either be #vars or 0
    z = [int(FILE.readline()) for i in range(4)]
    if read_vbtol:
        ampl_options.append(float(FILE.readline()))

    data.ampl_options = ampl_options
    return z


def _parse_objno_and_exitcode(FILE: io.TextIOBase, data: ASLSolFileData) -> None:
    line = FILE.readline().strip()
    objno = line.split(maxsplit=2)
    if not objno or objno[0] != 'objno':
        raise SolverError(
            f"Error reading `sol` file: expected 'objno'; received {line!r}."
        )
    elif len(objno) != 3:
        # TBD: [JDS, 10/2025] there are paths where writesol.c will
        # generate `objno` lines that contain only the objective number
        # and not the solve_code.  It is not clear to me that we should
        # generate an exception here.
        raise SolverError(
            "Error reading `sol` file: expected two numbers in 'objno' line; "
            f"received {line!r}."
        )
    data.objno = int(objno[1])
    data.solve_code = int(objno[2])


def _parse_suffixes(FILE: io.TextIOBase, data: ASLSolFileData) -> None:
    while line := FILE.readline():
        line = line.strip()
        if not line:
            continue

        line = line.split(maxsplit=6)
        if line[0] != 'suffix':
            # We assume this is the start of a section (like
            # kestrel_option) that comes *after* all suffixes.  We
            # will capture it (and everything after it) and return
            # it as a single "unparsed" text string.
            data.unparsed = ' '.join(line) + "\n" + ''.join(FILE)
            break

        # Each suffix is introduced by:
        #
        #   'suffix' <kind> <n> <namelen> <tablelen> <tablines>
        #   <sufname>
        #
        # Where:
        #   kind (int): bitmask indicating suffix data type and target
        #   n (int): number of values returned
        #   namelen (int): suffix name string length (including NULL termination)
        #   tablelen (int): length of the "table" string (including NULL)
        #   tablines (int): number of lines in the table
        #   sufname (str): suffix name
        kind = int(line[1])
        value_converter = float if kind & 4 else int
        suffix_target = kind & 3  # 0-var, 1-con, 2-obj, 3-prob

        num_values = int(line[2])
        # Note: we will use namelen to strip off the newline instead of
        # strip() in case the suffix name actually ended with whitespace
        # (evil, but technically allowed by the NL spec)
        suffix_name = FILE.readline()[: int(line[3]) - 1]

        # If the Suffix includes a value <-> string table, parse it.
        # The table should be a series of <tablines> lines of the form:
        #
        #     <val> <str> <description>
        #
        # The string representation of a suffix value is the row in the
        # table whose <val> is the largest value less than or equal to
        # the suffix value.  The table should be ordered by <val>.
        if int(line[4]):
            data.suffix_table[suffix_target, suffix_name] = [
                FILE.readline().strip().split(maxsplit=2) for _ in range(int(line[5]))
            ]
            for entry in data.suffix_table[suffix_target, suffix_name]:
                entry[0] = value_converter(entry[0])

        # Parse the actual suffix values
        if suffix_target == 0:  # Var
            data.var_suffixes[suffix_name] = suffix = {}
        elif suffix_target == 1:  # Con
            data.con_suffixes[suffix_name] = suffix = {}
        elif suffix_target == 2:  # Obj
            data.obj_suffixes[suffix_name] = suffix = {}
        elif suffix_target == 3:  # Prob
            suffix = {}
        # else:  # Unreachable: kind & 3 can ONLY be 0..3

        for cnt in range(num_values):
            suf_line = FILE.readline().split(maxsplit=1)
            suffix[int(suf_line[0])] = value_converter(suf_line[1])

        if suffix_target == 3 and suffix:
            assert len(suffix) == 1
            data.problem_suffixes[suffix_name] = next(iter(suffix.values()))

    return



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os.path

from pyomo.dataportal import TableData

# from pyomo.dataportal.plugins.db_table import (
#     pyodbc_available, pyodbc_db_Table, pypyodbc_available, pypyodbc_db_Table
# )
from pyomo.dataportal.factory import DataManagerFactory
from pyomo.common.errors import ApplicationError
from pyomo.common.dependencies import attempt_import, importlib, pyutilib


def _spreadsheet_importer():
    # verify pyutilib imported correctly the first time
    pyutilib.component
    return importlib.import_module('pyutilib.excel.spreadsheet')


spreadsheet, spreadsheet_available = attempt_import(
    'pyutilib.excel.spreadsheet', importer=_spreadsheet_importer
)


def _attempt_open_excel():
    if _attempt_open_excel.result is None:
        from pyutilib.excel.spreadsheet_win32com import ExcelSpreadsheet_win32com

        try:
            tmp = ExcelSpreadsheet_win32com()
            tmp._excel_dispatch()
            tmp._excel_quit()
            _attempt_open_excel.result = True
        except:
            _attempt_open_excel.result = False
    return _attempt_open_excel.result


_attempt_open_excel.result = None


class SheetTable(TableData):
    def __init__(self, ctype=None):
        TableData.__init__(self)
        self.ctype = ctype

    def open(self):
        if self.filename is None:
            raise IOError("No filename specified")
        if not os.path.exists(self.filename):
            raise IOError("Cannot find file '%s'" % self.filename)
        self.sheet = None
        if self._data is not None:
            self.sheet = self._data
        else:
            try:
                self.sheet = spreadsheet.ExcelSpreadsheet(
                    self.filename, ctype=self.ctype
                )
            except ApplicationError:
                raise

    def read(self):
        if self.sheet is None:
            return
        tmp = self.sheet.get_range(self.options.range, raw=True)
        if type(tmp) is float or type(tmp) is int:
            if not self.options.param is None:
                self._info = ["param"] + list(self.options.param) + [":=", tmp]
            elif len(self.options.symbol_map) == 1:
                self._info = [
                    "param",
                    self.options.symbol_map[self.options.symbol_map.keys()[0]],
                    ":=",
                    tmp,
                ]
            else:
                raise IOError(
                    "Data looks like a parameter, but multiple parameter names have been specified: %s"
                    % str(self.options.symbol_map)
                )
        elif len(tmp) == 0:
            raise IOError("Empty range '%s'" % self.options.range)
        else:
            if type(tmp[1]) in (list, tuple):
                tmp_ = tmp[1:]
            else:
                tmp_ = [[x] for x in tmp[1:]]
            self._set_data(tmp[0], tmp_)

    def close(self):
        if self._data is None and not self.sheet is None:
            del self.sheet


@DataManagerFactory.register("xls", "Excel XLS file interface")
class SheetTable_xls(SheetTable):
    def __init__(self):
        if spreadsheet.Interfaces()['win32com'].available and _attempt_open_excel():
            SheetTable.__init__(self, ctype='win32com')
        elif spreadsheet.Interfaces()['xlrd'].available:
            SheetTable.__init__(self, ctype='xlrd')
        else:
            raise RuntimeError(
                "No excel interface is available; install %s" % self.requirements()
            )

    def available(self):
        _inter = spreadsheet.Interfaces()
        return (_inter['win32com'].available and _attempt_open_excel()) or _inter[
            'xlrd'
        ].available

    def requirements(self):
        return "win32com or xlrd"


# @DataManagerFactory.register("xls", "Excel XLS file interface")
# class pyodbc_xls(pyodbc_db_base):

#     def __init__(self):
#         pyodbc_db_base.__init__(self)

#     def requirements(self):
#         return "pyodbc or pypyodbc"

#     def open(self):
#         if self.filename is None:
#             raise IOError("No filename specified")
#         if not os.path.exists(self.filename):
#             raise IOError("Cannot find file '%s'" % self.filename)
#         return pyodbc_db_base.open(self)


@DataManagerFactory.register("xlsx", "Excel XLSX file interface")
class SheetTable_xlsx(SheetTable):
    def __init__(self):
        if spreadsheet.Interfaces()['win32com'].available and _attempt_open_excel():
            SheetTable.__init__(self, ctype='win32com')
        elif spreadsheet.Interfaces()['openpyxl'].available:
            SheetTable.__init__(self, ctype='openpyxl')
        else:
            raise RuntimeError(
                "No excel interface is available; install %s" % self.requirements()
            )

    def available(self):
        _inter = spreadsheet.Interfaces()
        return (_inter['win32com'].available and _attempt_open_excel()) or _inter[
            'openpyxl'
        ].available

    def requirements(self):
        return "win32com or openpyxl"


#
# This class is OK, but the pyodbc interface doesn't work right now.
#

# @DataManagerFactory.register("xlsx", "Excel XLSX file interface")
# class SheetTable_xlsx(pyodbc_db_base):
#
#     def __init__(self):
#         pyodbc_db_base.__init__(self)
#
#     def requirements(self):
#         return "pyodbc or pypyodbc"
#
#     def open(self):
#         if self.filename is None:
#             raise IOError("No filename specified")
#         if not os.path.exists(self.filename):
#             raise IOError("Cannot find file '%s'" % self.filename)
#         return pyodbc_db_base.open(self)


# @DataManagerFactory.register("xlsb", "Excel XLSB file interface")
# class SheetTable_xlsb(pyodbc_db_base):
#
#     def __init__(self):
#         pyodbc_db_base.__init__(self)
#
#     def requirements(self):
#         return "pyodbc or pypyodbc"
#
#     def open(self):
#         if self.filename is None:
#             raise IOError("No filename specified")
#         if not os.path.exists(self.filename):
#             raise IOError("Cannot find file '%s'" % self.filename)
#         return pyodbc_db_base.open(self)


@DataManagerFactory.register("xlsm", "Excel XLSM file interface")
class SheetTable_xlsm(SheetTable):
    def __init__(self):
        if spreadsheet.Interfaces()['win32com'].available and _attempt_open_excel():
            SheetTable.__init__(self, ctype='win32com')
        elif spreadsheet.Interfaces()['openpyxl'].available:
            SheetTable.__init__(self, ctype='openpyxl')
        else:
            raise RuntimeError(
                "No excel interface is available; install %s" % self.requirements()
            )

    def available(self):
        _inter = spreadsheet.Interfaces()
        return (_inter['win32com'].available and _attempt_open_excel()) or _inter[
            'openpyxl'
        ].available

    def requirements(self):
        return "win32com or openpyxl"


# @DataManagerFactory.register("xlsm", "Excel XLSM file interface")
# class SheetTable_xlsm(pyodbc_db_base):
#
#     def __init__(self):
#         pyodbc_db_base.__init__(self)
#
#     def requirements(self):
#         return "pyodbc or pypyodbc"
#
#     def open(self):
#         if self.filename is None:
#             raise IOError("No filename specified")
#         if not os.path.exists(self.filename):
#             raise IOError("Cannot find file '%s'" % self.filename)
#         return pyodbc_db_base.open(self)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.kernel.base import ICategorizedObjectContainer


class IHomogeneousContainer(ICategorizedObjectContainer):
    """
    A partial implementation of the ICategorizedObjectContainer
    interface for implementations that store a single
    category of objects and that uses the same category
    as the objects it stores.

    Complete implementations need to set the _ctype
    attribute and declare the remaining required abstract
    properties of the ICategorizedObjectContainer base
    class.

    Note that this implementation allows nested storage of
    other :class:`ICategorizedObjectContainer`
    implementations that are defined with the same ctype.
    """

    __slots__ = ()

    #
    # Define the ICategorizedObjectContainer abstract methods
    #

    # def child(self, *args, **kwds):
    # ... not defined here

    # def children(self, *args, **kwds):
    # ... not defined here

    def components(self, active=True):
        """
        Generates an efficient traversal of all components
        stored under this container. Components are
        categorized objects that are either (1) not
        containers, or (2) are heterogeneous containers.

        Args:
            active (:const:`True`/:const:`None`): Controls
                whether or not to filter the iteration to
                include only the active part of the storage
                tree. The default is :const:`True`. Setting
                this keyword to :const:`None` causes the
                active status of objects to be ignored.

        Returns:
            iterator of components in the storage tree
        """
        assert active in (None, True)

        # if not active, then no children can be active
        if (active is not None) and (not self.active):
            return

        for child in self.children():
            if (active is None) or child.active:
                if (not child._is_container) or child._is_heterogeneous_container:
                    yield child
                else:
                    assert child._is_container
                    for obj in child.components(active=active):
                        yield obj
